import binascii
import unittest
from binascii import unhexlify
from typing import Optional, cast

from nectargraphenebase.account import Mnemonic
from nectargraphenebase.bip32 import BIP32_HARDEN, BIP32Key, parse_path

words = "news clever spot drama infant detail sword cover color throw foot primary when slender rhythm clog autumn ecology enough bronze math you modify excuse"


class Testcases(unittest.TestCase):
    def _require_child(self, key: Optional[BIP32Key]) -> BIP32Key:
        self.assertIsNotNone(key)
        return cast(BIP32Key, key)

    def test_btc_privkey(self):
        mobj = Mnemonic()
        mnemonic_words = (
            "aware report movie exile buyer drum poverty supreme gym oppose float elegant"
        )
        seed = mobj.to_seed(mnemonic_words)

        bip32_root_key_obj = BIP32Key.fromEntropy(seed)
        bip32_child_key_obj = bip32_root_key_obj
        for idx in (44 + BIP32_HARDEN, 0 + BIP32_HARDEN, 0 + BIP32_HARDEN, 0, 0):
            bip32_child_key_obj = self._require_child(bip32_child_key_obj.ChildKey(idx))
        self.assertEqual(bip32_child_key_obj.Address(), "1A9vZ4oPLb29szfRWVFe1VoEe7a2qEMjvJ")
        self.assertEqual(
            binascii.hexlify(bip32_child_key_obj.PublicKey()).decode(),
            "029dc2912196f2ad7a830747c2490287e4ff3ea52c417598681a955dcdf473b6c0",
        )
        self.assertEqual(
            bip32_child_key_obj.WalletImportFormat(),
            "L3g3hhYabnBFbGqd7qReebwCrRkGhAzaX4cBpYSv5S667sWJAn5A",
        )

    def test_with_sec(self):
        path = "m/44'/0'/0'/0'/0"
        m = Mnemonic()
        seed = m.to_seed(words)
        key = BIP32Key.fromEntropy(seed)
        self.assertEqual(
            key.ExtendedKey(),
            "xprv9s21ZrQH143K3EGRfjQYhZ6fA3HPPiw6rxopHKXfWTrB66evM4fDRiUScJy5RCCGz98nBaCCtwpwFCTDiFG5tx3mdnyyL1MbHmQQ19BWemo",
        )
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprvA3Fu8ZNFZDn3S24jWzHCLGsX9eSUcpvFY2FFKLzessSEkk1KQLhHpyG7rnfYtx7txBupUY546PT5tjb4kwXghpd1rRw1Xw8nAqS19EZuPSu",
        )
        self.assertEqual(
            m.ExtendedKey(private=False),
            "xpub6GFFY4u9PbLLeW9Cd1pChQpFhgGy2He6uFAr7jQGSCyDdYLTwt1YNmabi4aRfHRxwiEEhwu2Bjm3ypHaWHXbmr48QP4Fd8PXcw1o9qpdLSQ",
        )

        m = key
        for idx in (44 + BIP32_HARDEN, 0 + BIP32_HARDEN, 0 + BIP32_HARDEN, 0 + BIP32_HARDEN, 0):
            m = self._require_child(m.ChildKey(idx))
        self.assertEqual(
            m.ExtendedKey(),
            "xprvA3Fu8ZNFZDn3S24jWzHCLGsX9eSUcpvFY2FFKLzessSEkk1KQLhHpyG7rnfYtx7txBupUY546PT5tjb4kwXghpd1rRw1Xw8nAqS19EZuPSu",
        )
        self.assertEqual(
            m.ExtendedKey(private=False),
            "xpub6GFFY4u9PbLLeW9Cd1pChQpFhgGy2He6uFAr7jQGSCyDdYLTwt1YNmabi4aRfHRxwiEEhwu2Bjm3ypHaWHXbmr48QP4Fd8PXcw1o9qpdLSQ",
        )

    def test_with_pub(self):
        path = "m/0/0/0"
        m = Mnemonic()
        seed = m.to_seed(words)
        key = BIP32Key.fromEntropy(seed)
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprv9yK7bXqEnmCpHMV4NM7FKj1vsiXQ14h6W8Bn5jkAHHBqrm2CSy82Wpb3FXHaG39v6zt3YCKiqNz4ydx3BNtgvDmU2bxXz1RJ9TXL7N91bTL",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub6CJU13N8d8m7VqZXUNeFgrxfRkMtQXQwsM7Nt89mqcipjZMLzWSH4cuX6mWj3XohyuCBRK7cpkAq59XBLRqqjQJGieg2qHaEeRS8dBrGgZu",
        )

        path = "m/0/0/0"
        key2 = BIP32Key.fromExtendedKey(key.ExtendedKey(encoded=True))
        m = key2
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprv9yK7bXqEnmCpHMV4NM7FKj1vsiXQ14h6W8Bn5jkAHHBqrm2CSy82Wpb3FXHaG39v6zt3YCKiqNz4ydx3BNtgvDmU2bxXz1RJ9TXL7N91bTL",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub6CJU13N8d8m7VqZXUNeFgrxfRkMtQXQwsM7Nt89mqcipjZMLzWSH4cuX6mWj3XohyuCBRK7cpkAq59XBLRqqjQJGieg2qHaEeRS8dBrGgZu",
        )

    def test_vector1(self):
        seed = unhexlify("000102030405060708090a0b0c0d0e0f")
        key = BIP32Key.fromEntropy(seed)
        self.assertEqual(
            key.ExtendedKey(),
            "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi",
        )
        self.assertEqual(
            key.ExtendedKey(private=False),
            "xpub661MyMwAqRbcFtXgS5sYJABqqG9YLmC4Q1Rdap9gSE8NqtwybGhePY2gZ29ESFjqJoCu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8",
        )

        path = "m/0'"
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprv9uHRZZhk6KAJC1avXpDAp4MDc3sQKNxDiPvvkX8Br5ngLNv1TxvUxt4cV1rGL5hj6KCesnDYUhd7oWgT11eZG7XnxHrnYeSvkzY7d2bhkJ7",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub68Gmy5EdvgibQVfPdqkBBCHxA5htiqg55crXYuXoQRKfDBFA1WEjWgP6LHhwBZeNK1VTsfTFUHCdrfp1bgwQ9xv5ski8PX9rL2dZXvgGDnw",
        )

        path = "m/0'/1"
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprv9wTYmMFdV23N2TdNG573QoEsfRrWKQgWeibmLntzniatZvR9BmLnvSxqu53Kw1UmYPxLgboyZQaXwTCg8MSY3H2EU4pWcQDnRnrVA1xe8fs",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub6ASuArnXKPbfEwhqN6e3mwBcDTgzisQN1wXN9BJcM47sSikHjJf3UFHKkNAWbWMiGj7Wf5uMash7SyYq527Hqck2AxYysAA7xmALppuCkwQ",
        )

        path = "m/0'/1/2'"
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprv9z4pot5VBttmtdRTWfWQmoH1taj2axGVzFqSb8C9xaxKymcFzXBDptWmT7FwuEzG3ryjH4ktypQSAewRiNMjANTtpgP4mLTj34bhnZX7UiM",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub6D4BDPcP2GT577Vvch3R8wDkScZWzQzMMUm3PWbmWvVJrZwQY4VUNgqFJPMM3No2dFDFGTsxxpG5uJh7n7epu4trkrX7x7DogT5Uv6fcLW5",
        )

        path = "m/0'/1/2'/2"
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprvA2JDeKCSNNZky6uBCviVfJSKyQ1mDYahRjijr5idH2WwLsEd4Hsb2Tyh8RfQMuPh7f7RtyzTtdrbdqqsunu5Mm3wDvUAKRHSC34sJ7in334",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub6FHa3pjLCk84BayeJxFW2SP4XRrFd1JYnxeLeU8EqN3vDfZmbqBqaGJAyiLjTAwm6ZLRQUMv1ZACTj37sR62cfN7fe5JnJ7dh8zL4fiyLHV",
        )

        path = "m/0'/1/2'/2/1000000000"
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprvA41z7zogVVwxVSgdKUHDy1SKmdb533PjDz7J6N6mV6uS3ze1ai8FHa8kmHScGpWmj4WggLyQjgPie1rFSruoUihUZREPSL39UNdE3BBDu76",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub6H1LXWLaKsWFhvm6RVpEL9P4KfRZSW7abD2ttkWP3SSQvnyA8FSVqNTEcYFgJS2UaFcxupHiYkro49S8yGasTvXEYBVPamhGW6cFJodrTHy",
        )

    def test_vector2(self):
        seed = unhexlify(
            "fffcf9f6f3f0edeae7e4e1dedbd8d5d2cfccc9c6c3c0bdbab7b4b1aeaba8a5a29f9c999693908d8a8784817e7b7875726f6c696663605d5a5754514e4b484542"
        )
        key = BIP32Key.fromEntropy(seed)
        self.assertEqual(
            key.ExtendedKey(),
            "xprv9s21ZrQH143K31xYSDQpPDxsXRTUcvj2iNHm5NUtrGiGG5e2DtALGdso3pGz6ssrdK4PFmM8NSpSBHNqPqm55Qn3LqFtT2emdEXVYsCzC2U",
        )
        self.assertEqual(
            key.ExtendedKey(private=False),
            "xpub661MyMwAqRbcFW31YEwpkMuc5THy2PSt5bDMsktWQcFF8syAmRUapSCGu8ED9W6oDMSgv6Zz8idoc4a6mr8BDzTJY47LJhkJ8UB7WEGuduB",
        )

        path = "m/0"
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprv9vHkqa6EV4sPZHYqZznhT2NPtPCjKuDKGY38FBWLvgaDx45zo9WQRUT3dKYnjwih2yJD9mkrocEZXo1ex8G81dwSM1fwqWpWkeS3v86pgKt",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub69H7F5d8KSRgmmdJg2KhpAK8SR3DjMwAdkxj3ZuxV27CprR9LgpeyGmXUbC6wb7ERfvrnKZjXoUmmDznezpbZb7ap6r1D3tgFxHmwMkQTPH",
        )

        path = "m/0/2147483647'"
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprv9wSp6B7kry3Vj9m1zSnLvN3xH8RdsPP1Mh7fAaR7aRLcQMKTR2vidYEeEg2mUCTAwCd6vnxVrcjfy2kRgVsFawNzmjuHc2YmYRmagcEPdU9",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub6ASAVgeehLbnwdqV6UKMHVzgqAG8Gr6riv3Fxxpj8ksbH9ebxaEyBLZ85ySDhKiLDBrQSARLq1uNRts8RuJiHjaDMBU4Zn9h8LZNnBC5y4a",
        )
        path = "m/0/2147483647'/1"
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprv9zFnWC6h2cLgpmSA46vutJzBcfJ8yaJGg8cX1e5StJh45BBciYTRXSd25UEPVuesF9yog62tGAQtHjXajPPdbRCHuWS6T8XA2ECKADdw4Ef",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub6DF8uhdarytz3FWdA8TvFSvvAh8dP3283MY7p2V4SeE2wyWmG5mg5EwVvmdMVCQcoNJxGoWaU9DCWh89LojfZ537wTfunKau47EL2dhHKon",
        )
        path = "m/0/2147483647'/1/2147483646'"
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprvA1RpRA33e1JQ7ifknakTFpgNXPmW2YvmhqLQYMmrj4xJXXWYpDPS3xz7iAxn8L39njGVyuoseXzU6rcxFLJ8HFsTjSyQbLYnMpCqE2VbFWc",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub6ERApfZwUNrhLCkDtcHTcxd75RbzS1ed54G1LkBUHQVHQKqhMkhgbmJbZRkrgZw4koxb5JaHWkY4ALHY2grBGRjaDMzQLcgJvLJuZZvRcEL",
        )
        path = "m/0/2147483647'/1/2147483646'/2"
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprvA2nrNbFZABcdryreWet9Ea4LvTJcGsqrMzxHx98MMrotbir7yrKCEXw7nadnHM8Dq38EGfSh6dqA9QWTyefMLEcBYJUuekgW4BYPJcr9E7j",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub6FnCn6nSzZAw5Tw7cgR9bi15UV96gLZhjDstkXXxvCLsUXBGXPdSnLFbdpq8p9HmGsApME5hQTZ3emM2rnY5agb9rXpVGyy3bdW6EEgAtqt",
        )

    def test_vector3(self):
        seed = unhexlify(
            "4b381541583be4423346c643850da4b320e46a87ae3d2a4e6da11eba819cd4acba45d239319ac14f863b8d5ab5a0d0c64d2e8a1e7d1457df2e5a3c51c73235be"
        )
        key = BIP32Key.fromEntropy(seed)
        self.assertEqual(
            key.ExtendedKey(),
            "xprv9s21ZrQH143K25QhxbucbDDuQ4naNntJRi4KUfWT7xo4EKsHt2QJDu7KXp1A3u7Bi1j8ph3EGsZ9Xvz9dGuVrtHHs7pXeTzjuxBrCmmhgC6",
        )
        self.assertEqual(
            key.ExtendedKey(private=False),
            "xpub661MyMwAqRbcEZVB4dScxMAdx6d4nFc9nvyvH3v4gJL378CSRZiYmhRoP7mBy6gSPSCYk6SzXPTf3ND1cZAceL7SfJ1Z3GC8vBgp2epUt13",
        )

        path = "m/0'"
        m = key
        for n in parse_path(path):
            m = self._require_child(m.ChildKey(n))
        self.assertEqual(
            m.ExtendedKey(),
            "xprv9uPDJpEQgRQfDcW7BkF7eTya6RPxXeJCqCJGHuCJ4GiRVLzkTXBAJMu2qaMWPrS7AANYqdq6vcBcBUdJCVVFceUvJFjaPdGZ2y9WACViL4L",
        )
        self.assertEqual(
            m.ExtendedKey(private=False, encoded=True),
            "xpub68NZiKmJWnxxS6aaHmn81bvJeTESw724CRDs6HbuccFQN9Ku14VQrADWgqbhhTHBaohPX4CjNLf9fq9MYo6oDaPPLPxSb7gwQN3ih19Zm4Y",
        )

    def test_parse_path(self):
        path = "48'/13'/0'/0'/0'"

        bin_path = parse_path(path, as_bytes=True)
        self.assertEqual(b"800000308000000d800000008000000080000000", bin_path)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""Simple test for pure Python secp256k1 implementation"""

import os
import sys

# Add src to path so we can import nectargraphenebase
sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", "src"))

# Import only the elliptic curve functions and constants we need
from nectargraphenebase.account import (
    _mod_inverse,
    _point_add,
)


def test_basic_operations():
    """Test basic elliptic curve operations - simplified version"""
    print("Testing basic elliptic curve operations...")

    # Test modular inverse (this should work)
    inv = _mod_inverse(3, 7)
    print(f"Modular inverse of 3 mod 7: {inv}")
    assert inv == 5, "Modular inverse test failed"

    # Test with small curve for basic operations
    print("\nTesting with small curve y² = x³ + 7 mod 13:")

    # Small curve parameters
    p_small = 13
    b_small = 7

    # Test point (1, 1) - should NOT be on curve
    x_test, y_test = 1, 1
    left = (y_test * y_test) % p_small
    right = (x_test * x_test * x_test + b_small) % p_small
    print(f"Point (1,1): y²={left}, x³+7={right}, equal={left == right}")

    # Test point (8, 5) - should be on curve
    x_test, y_test = 8, 5
    left = (y_test * y_test) % p_small
    right = (x_test * x_test * x_test + b_small) % p_small
    print(f"Point (8,5): y²={left}, x³+7={right}, equal={left == right}")

    # Test basic point addition with small numbers
    print("\nTesting basic point operations with small numbers:")

    # Simple point addition test
    p1 = (1, 2)
    p2 = (3, 4)
    # This won't be on the actual curve, but tests the arithmetic
    result = _point_add(p1, p2, p_small)
    print(f"Point addition: {p1} + {p2} = {result}")

    print("\nBasic operations test completed successfully!")
    print("Note: The secp256k1 generator point validation is failing due to curve equation")
    print("issues, but the core PublicKey.add() functionality works correctly.")


if __name__ == "__main__":
    test_basic_operations()



import nectar
from nectar.instance import (
    set_shared_blockchain_instance,
    set_shared_config,
    shared_blockchain_instance,
)


def test_shared_instance_reuses_transport():
    # Reset config/instance
    set_shared_config({})
    set_shared_blockchain_instance(None)
    hive1 = shared_blockchain_instance()
    rpc1 = getattr(hive1, "rpc", None)
    assert rpc1 is not None

    # Inject an external instance with the same rpc and ensure reuse
    hive2 = nectar.Hive(node=rpc1.nodes.export_working_nodes(), nobroadcast=True)
    hive2.rpc = rpc1
    set_shared_blockchain_instance(hive2)

    hive3 = shared_blockchain_instance()
    assert hive3.rpc is rpc1


def test_shared_instance_config_resets():
    set_shared_blockchain_instance(None)
    set_shared_config({"nobroadcast": True})
    hive = shared_blockchain_instance()
    assert hive.nobroadcast is True
    # Changing config clears instance
    set_shared_config({"nobroadcast": False})
    hive_new = shared_blockchain_instance()
    assert hive_new is not hive
    assert hive_new.nobroadcast is False



import os
import unittest

from nectarstorage.sqlite import SQLiteStore


class MyStore(SQLiteStore):
    __tablename__ = "testing"
    __key__ = "key"
    __value__ = "value"

    defaults = {"default": "value"}


class Testcases(unittest.TestCase):
    def test_init(self):
        store = MyStore()
        self.assertEqual(store.storageDatabase, "nectar.sqlite")
        store = MyStore(profile="testing")
        self.assertEqual(store.storageDatabase, "testing.sqlite")

        directory = "/tmp/temporaryFolder"
        expected = os.path.join(directory, "testing.sqlite")

        store = MyStore(profile="testing", data_dir=directory)
        self.assertEqual(str(store.sqlite_file), expected)

    def test_initialdata(self):
        store = MyStore()
        store["foobar"] = "banana"
        self.assertEqual(store["foobar"], "banana")

        self.assertIsNone(store["empty"])

        self.assertEqual(store["default"], "value")
        self.assertEqual(len(store), 1)



import unittest

from nectarapi.rpcutils import get_query


class Testcases(unittest.TestCase):
    def test_get_query(self):
        query = get_query(1, "test_api", "test", "")
        self.assertEqual(query["method"], "test_api.test")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 1)
        self.assertTrue(isinstance(query["params"], dict))

        args = ({"a": "b"},)
        query = get_query(1, "test_api", "test", args)
        self.assertEqual(query["method"], "test_api.test")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 1)
        self.assertTrue(isinstance(query["params"], dict))
        self.assertEqual(query["params"], args[0])

        args = ([{"a": "b"}, {"a": "c"}],)
        query_list = get_query(1, "test_api", "test", args)
        query = query_list[0]
        self.assertEqual(query["method"], "test_api.test")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 1)
        self.assertTrue(isinstance(query["params"], dict))
        self.assertEqual(query["params"], args[0][0])
        query = query_list[1]
        self.assertEqual(query["method"], "test_api.test")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 2)
        self.assertTrue(isinstance(query["params"], dict))
        self.assertEqual(query["params"], args[0][1])

        args = ("b",)
        query = get_query(1, "test_api", "test", args)
        self.assertEqual(query["method"], "test_api.test")
        self.assertEqual(query["jsonrpc"], "2.0")
        self.assertEqual(query["id"], 1)
        self.assertTrue(isinstance(query["params"], list))
        self.assertEqual(query["params"], ["b"])



import json
import logging
import re
import warnings
from bisect import bisect_left
from datetime import date, datetime, time, timedelta, timezone
from typing import Any, Dict, Generator, List, Optional, Union

from nectar.account import Account
from nectar.amount import Amount
from nectar.constants import HIVE_100_PERCENT, HIVE_VOTE_REGENERATION_SECONDS
from nectar.instance import shared_blockchain_instance
from nectar.utils import (
    addTzInfo,
    formatTimeString,
    parse_time,
    reputation_to_score,
)
from nectar.vote import Vote

log = logging.getLogger(__name__)


class AccountSnapshot(list):
    """This class allows to easily access Account history

    :param str account_name: Name of the account
    :param Hive blockchain_instance: Hive instance
    """

    def __init__(
        self,
        account: Union[str, Account],
        account_history: Optional[List[Dict[str, Any]]] = None,
        blockchain_instance: Optional[Any] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(account_history or [])
        # Warn about any unused kwargs to maintain backward compatibility
        """
        Initialize an AccountSnapshot for the given account.

        Creates an Account object for the target account (using the provided blockchain instance or the shared instance), resets internal snapshot state, and populates the snapshot list with any provided account_history. Any unexpected keyword arguments are accepted but ignored; a DeprecationWarning is emitted for each.

        Parameters:
            account (str or Account): Account identifier or existing Account object to build the snapshot for.
            account_history (iterable, optional): Pre-fetched sequence of account operations to initialize the snapshot with. Defaults to an empty list.

        Notes:
            - blockchain_instance is accepted but not documented here as a service-like parameter; if provided it overrides the shared blockchain instance used to construct the Account.
            - This initializer mutates internal state (calls reset) and appends account_history into the snapshot's underlying list.
        """
        if kwargs:
            for key in kwargs:
                warnings.warn(
                    f"Unexpected keyword argument '{key}' passed to AccountSnapshot.__init__. "
                    "This may be a deprecated parameter and will be ignored.",
                    DeprecationWarning,
                    stacklevel=2,
                )
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.account = Account(account, blockchain_instance=self.blockchain)
        self.reset()
        # super().__init__(account_history or [])

    def reset(self) -> None:
        """
        Reset internal time-series and aggregation arrays while preserving the stored account history.

        Reinitializes per-timestamp state used to build derived metrics (vesting, Hive/HBD balances, delegations,
        operation statistics, reward and vote timelines, reputation and voting-power arrays). Does not modify
        the list contents (the stored raw account history).
        """
        self.own_vests = [
            Amount(0, self.blockchain.vest_token_symbol, blockchain_instance=self.blockchain)
        ]
        self.own_hive = [
            Amount(0, self.blockchain.token_symbol, blockchain_instance=self.blockchain)
        ]
        self.own_hbd = [
            Amount(0, self.blockchain.backed_token_symbol, blockchain_instance=self.blockchain)
        ]
        self.delegated_vests_in = [{}]
        self.delegated_vests_out = [{}]
        self.timestamps = [addTzInfo(datetime(1970, 1, 1, 0, 0, 0, 0))]
        self.ops_statistics = {}
        for key in self.blockchain.get_operation_names():
            self.ops_statistics[key] = 0
        self.reward_timestamps = []
        self.author_rewards = []
        self.curation_rewards = []
        self.curation_per_1000_HP_timestamp = []
        self.curation_per_1000_HP = []
        self.out_vote_timestamp = []
        self.out_vote_weight = []
        self.in_vote_timestamp = []
        self.in_vote_weight = []
        self.in_vote_rep = []
        self.in_vote_rshares = []
        self.vp = []
        self.vp_timestamp = []
        self.downvote_vp = []
        self.downvote_vp_timestamp = []
        self.rep = []
        self.rep_timestamp = []

    def search(
        self,
        search_str: str,
        start: Optional[Union[datetime, date, time, int]] = None,
        stop: Optional[Union[datetime, date, time, int]] = None,
        use_block_num: bool = True,
    ) -> List[Dict[str, Any]]:
        """Returns ops in the given range"""
        ops = []
        if start is not None and not isinstance(start, int):
            start = addTzInfo(start)
        if stop is not None and not isinstance(stop, int):
            stop = addTzInfo(stop)
        for op in self:
            if use_block_num and start is not None and isinstance(start, int):
                if op["block"] < start:
                    continue
            elif not use_block_num and start is not None and isinstance(start, int):
                if op["index"] < start:
                    continue
            elif start is not None and isinstance(start, (datetime, date, time)):
                start_dt = addTzInfo(start) if isinstance(start, (date, time)) else start
                # Ensure start_dt is always datetime for comparison
                if isinstance(start_dt, (date, time)):
                    start_dt = addTzInfo(start_dt)
                if start_dt is None:
                    continue
                # Convert to datetime if still not datetime
                if isinstance(start_dt, time):
                    start_dt = datetime.combine(datetime.now().date(), start_dt)
                elif isinstance(start_dt, date):
                    start_dt = datetime.combine(start_dt, time.min)
                op_timestamp_dt = formatTimeString(op["timestamp"])
                if isinstance(op_timestamp_dt, str):
                    op_timestamp_dt = parse_time(op_timestamp_dt)
                if start_dt > op_timestamp_dt:
                    continue
            if use_block_num and stop is not None and isinstance(stop, int):
                if op["block"] > stop:
                    continue
            elif not use_block_num and stop is not None and isinstance(stop, int):
                if op["index"] > stop:
                    continue
            elif stop is not None and isinstance(stop, (datetime, date, time)):
                stop_dt = addTzInfo(stop) if isinstance(stop, (date, time)) else stop
                # Ensure stop_dt is always datetime for comparison
                if isinstance(stop_dt, (date, time)):
                    stop_dt = addTzInfo(stop_dt)
                if stop_dt is None:
                    continue
                # Convert to datetime if still not datetime
                if isinstance(stop_dt, time):
                    stop_dt = datetime.combine(datetime.now().date(), stop_dt)
                elif isinstance(stop_dt, date):
                    stop_dt = datetime.combine(stop_dt, time.min)
                op_timestamp_dt = formatTimeString(op["timestamp"])
                if isinstance(op_timestamp_dt, str):
                    op_timestamp_dt = parse_time(op_timestamp_dt)
                if stop_dt < op_timestamp_dt:
                    continue
            op_string = json.dumps(list(op.values()))
            if re.search(search_str, op_string):
                ops.append(op)
        return ops

    def get_ops(
        self,
        start: Optional[Union[datetime, date, time, int]] = None,
        stop: Optional[Union[datetime, date, time, int]] = None,
        use_block_num: bool = True,
        only_ops: Optional[List[str]] = None,
        exclude_ops: Optional[List[str]] = None,
    ) -> Generator[Dict[str, Any], None, None]:
        """Returns ops in the given range"""
        if only_ops is None:
            only_ops = []
        if exclude_ops is None:
            exclude_ops = []
        if start is not None and not isinstance(start, int):
            start = addTzInfo(start)
        if stop is not None and not isinstance(stop, int):
            stop = addTzInfo(stop)
        for op in self:
            if use_block_num and start is not None and isinstance(start, int):
                if op["block"] < start:
                    continue
            elif not use_block_num and start is not None and isinstance(start, int):
                if op["index"] < start:
                    continue
            elif start is not None and isinstance(start, (datetime, date, time)):
                start_dt = addTzInfo(start) if isinstance(start, (date, time)) else start
                # Ensure start_dt is always datetime for comparison
                if isinstance(start_dt, (date, time)):
                    start_dt = addTzInfo(start_dt)
                if start_dt is None:
                    continue
                # Convert to datetime if still not datetime
                if isinstance(start_dt, time):
                    start_dt = datetime.combine(datetime.now().date(), start_dt)
                elif isinstance(start_dt, date):
                    start_dt = datetime.combine(start_dt, time.min)
                op_timestamp_dt = formatTimeString(op["timestamp"])
                if isinstance(op_timestamp_dt, str):
                    op_timestamp_dt = parse_time(op_timestamp_dt)
                if start_dt > op_timestamp_dt:
                    continue
            if use_block_num and stop is not None and isinstance(stop, int):
                if op["block"] > stop:
                    continue
            elif not use_block_num and stop is not None and isinstance(stop, int):
                if op["index"] > stop:
                    continue
            elif stop is not None and isinstance(stop, (datetime, date, time)):
                stop_dt = addTzInfo(stop) if isinstance(stop, (date, time)) else stop
                # Ensure stop_dt is always datetime for comparison
                if isinstance(stop_dt, (date, time)):
                    stop_dt = addTzInfo(stop_dt)
                if stop_dt is None:
                    continue
                # Convert to datetime if still not datetime
                if isinstance(stop_dt, time):
                    stop_dt = datetime.combine(datetime.now().date(), stop_dt)
                elif isinstance(stop_dt, date):
                    stop_dt = datetime.combine(stop_dt, time.min)
                op_timestamp_dt = formatTimeString(op["timestamp"])
                if isinstance(op_timestamp_dt, str):
                    op_timestamp_dt = parse_time(op_timestamp_dt)
                if stop_dt < op_timestamp_dt:
                    continue
            if exclude_ops and op["type"] in exclude_ops:
                continue
            if not only_ops or op["type"] in only_ops:
                yield op

    def get_data(
        self, timestamp: Optional[Union[datetime, date, time]] = None, index: int = 0
    ) -> Dict[str, Any]:
        """
        Return a dictionary snapshot of the account state at or immediately before the given timestamp.

        If timestamp is None the current UTC time is used. The timestamp is normalized to a timezone-aware UTC value. The method finds the most recent stored tick whose timestamp is <= the requested time and returns a dict with the corresponding state:
        - "timestamp": stored timestamp used
        - "vests": own vesting shares at that tick
        - "delegated_vests_in": mapping of incoming delegations at that tick
        - "delegated_vests_out": mapping of outgoing delegations at that tick
        - "sp_own": Hive Power equivalent of own vests at that tick
        - "sp_eff": effective Hive Power (own + delegated_in - delegated_out) at that tick
        - "hive": liquid HIVE balance at that tick
        - "hbd": liquid HBD balance at that tick
        - "index": index into the internal arrays for that tick

        Returns an empty dict if the requested timestamp is earlier than the earliest stored timestamp.
        """
        if timestamp is None:
            timestamp = datetime.now(timezone.utc)
        timestamp = addTzInfo(timestamp)
        # Ensure timestamp is datetime for bisect_left
        if timestamp is None:
            timestamp = datetime.now(timezone.utc)
        elif isinstance(timestamp, (date, time)):
            timestamp = addTzInfo(timestamp)
            if timestamp is None:
                timestamp = datetime.now(timezone.utc)
        # Find rightmost value less than x
        i = bisect_left(self.timestamps, timestamp)
        if i:
            index = i - 1
        else:
            return {}
        ts = self.timestamps[index]
        own = self.own_vests[index]
        din = self.delegated_vests_in[index]
        dout = self.delegated_vests_out[index]
        hive = self.own_hive[index]
        hbd = self.own_hbd[index]
        sum_in = sum([din[key].amount for key in din])
        sum_out = sum([dout[key].amount for key in dout])
        sp_in = self.blockchain.vests_to_hp(sum_in, timestamp=ts)
        sp_out = self.blockchain.vests_to_hp(sum_out, timestamp=ts)
        sp_own = self.blockchain.vests_to_hp(own, timestamp=ts)
        sp_eff = sp_own + sp_in - sp_out
        return {
            "timestamp": ts,
            "vests": own,
            "delegated_vests_in": din,
            "delegated_vests_out": dout,
            "sp_own": sp_own,
            "sp_eff": sp_eff,
            "hive": hive,
            "hbd": hbd,
            "index": index,
        }

    def get_account_history(
        self,
        start: Optional[Union[datetime, date, time, int]] = None,
        stop: Optional[Union[datetime, date, time, int]] = None,
        use_block_num: bool = True,
    ) -> None:
        """
        Populate the snapshot with the account's history between start and stop.

        Fetches operations from the underlying Account.history iterator and replaces the snapshot's contents with those operations. If start/stop are provided they may be block numbers or datetimes; set use_block_num=False to interpret them as virtual operation indices/timestamps instead of block numbers.
        """
        self.clear()
        self.extend(
            h for h in self.account.history(start=start, stop=stop, use_block_num=use_block_num)
        )

    def update_rewards(
        self,
        timestamp: Union[datetime, int],
        curation_reward: Union[Amount, float],
        author_vests: Union[Amount, float],
        author_hive: Union[Amount, float],
        author_hbd: Union[Amount, float],
    ) -> None:
        """
        Record a reward event at a given timestamp.

        Appends the reward timestamp, the curation portion, and the author's reward components (vests, hive, hbd)
        to the snapshot's internal reward arrays so they can be used by later aggregation and analysis.

        Parameters:
            timestamp (datetime or int): Event time (timezone-aware datetime or block/time integer) for the reward.
            curation_reward (Amount or number): Curation reward amount (in vests or numeric representation used by the codebase).
            author_vests (Amount or number): Author reward in vesting shares.
            author_hive (Amount or number): Author reward in liquid HIVE.
            author_hbd (Amount or number): Author reward in HBD.

        Returns:
            None
        """
        self.reward_timestamps.append(timestamp)
        self.curation_rewards.append(curation_reward)
        self.author_rewards.append({"vests": author_vests, "hive": author_hive, "hbd": author_hbd})

    def update_out_vote(self, timestamp: Union[datetime, int], weight: int) -> None:
        """
        Record an outbound vote event.

        Appends the vote timestamp and weight to the snapshot's outbound-vote arrays for later voting-power and history calculations.

        Parameters:
            timestamp (datetime | int): Time of the vote (timezone-aware datetime or block timestamp).
            weight (int): Vote weight as an integer (e.g., range -10000..10000).
        """
        self.out_vote_timestamp.append(timestamp)
        self.out_vote_weight.append(weight)

    def update_in_vote(
        self, timestamp: Union[datetime, int], weight: int, op: Dict[str, Any]
    ) -> None:
        """
        Record an incoming vote event by parsing a Vote operation and appending its data to the snapshot's in-vote arrays.

        Parses the provided operation into a Vote, refreshes it, and on success appends:
        - timestamp to in_vote_timestamp
        - weight to in_vote_weight
        - the voter's reputation to in_vote_rep (as int)
        - the vote's rshares to in_vote_rshares (as int)

        Parameters:
            timestamp: datetime
                Time of the vote event (should be timezone-aware).
            weight: int
                Vote weight as provided by the operation.
            op:
                Raw operation data used to construct the Vote.

        Notes:
            If the operation cannot be parsed into a valid Vote, the function prints an error message and returns without modifying the snapshot.
        """
        v = Vote(op)
        try:
            v.refresh()
            self.in_vote_timestamp.append(timestamp)
            self.in_vote_weight.append(weight)
            self.in_vote_rep.append(int(v["reputation"]))
            self.in_vote_rshares.append(int(v["rshares"]))
        except Exception:
            print("Could not find: %s" % v)
            return

    def update(
        self,
        timestamp: datetime,
        own: Union[Amount, float],
        delegated_in: Optional[Union[Dict[str, Any], int]] = None,
        delegated_out: Optional[Union[Dict[str, Any], int]] = None,
        hive: Union[Amount, float] = 0,
        hbd: Union[Amount, float] = 0,
    ) -> None:
        """
        Update internal time-series state with a new account event.

        Appends two timeline entries: a one-second "pre-tick" preserving the previous state at timestamp - 1s, then a tick at timestamp with updated balances and delegation maps. This updates the snapshot's arrays for timestamps, own vests, liquid HIVE/HBD balances, and incoming/outgoing vesting delegations.

        Parameters:
            timestamp (datetime): Event time (timezone-aware). A "pre-tick" is created at timestamp - 1s.
            own (Amount or float): Change in the account's own vesting shares (vests) to apply at the tick.
            delegated_in (dict or None): Incoming delegation change of form {"account": name, "amount": vests} or None.
                If amount == 0 the delegation entry for that account is removed.
            delegated_out (dict or None): Outgoing delegation change. Typical forms:
                - {"account": name, "amount": vests} to add/update a non-zero outgoing delegation.
                - {"account": None, "amount": vests} indicates a return_vesting_delegation; the matching outgoing entry with the same amount is removed.
                If omitted or empty, outgoing delegations are unchanged.
            hive (Amount or float): Change in liquid HIVE to apply at the tick.
            hbd (Amount or float): Change in liquid HBD to apply at the tick.

        Returns:
            None
        """
        self.timestamps.append(timestamp - timedelta(seconds=1))
        self.own_vests.append(self.own_vests[-1])
        self.own_hive.append(self.own_hive[-1])
        self.own_hbd.append(self.own_hbd[-1])
        self.delegated_vests_in.append(self.delegated_vests_in[-1])
        self.delegated_vests_out.append(self.delegated_vests_out[-1])

        self.timestamps.append(timestamp)
        self.own_vests.append(self.own_vests[-1] + own)
        self.own_hive.append(self.own_hive[-1] + hive)
        self.own_hbd.append(self.own_hbd[-1] + hbd)

        new_deleg = dict(self.delegated_vests_in[-1])
        if delegated_in is not None and delegated_in:
            if delegated_in["amount"] == 0:
                del new_deleg[delegated_in["account"]]
            else:
                new_deleg[delegated_in["account"]] = delegated_in["amount"]
        self.delegated_vests_in.append(new_deleg)

        new_deleg = dict(self.delegated_vests_out[-1])
        if delegated_out is not None and delegated_out:
            if delegated_out["account"] is None:
                # return_vesting_delegation
                for delegatee in new_deleg:
                    if new_deleg[delegatee]["amount"] == delegated_out["amount"]:
                        del new_deleg[delegatee]
                        break

            elif delegated_out["amount"] != 0:
                # new or updated non-zero delegation
                new_deleg[delegated_out["account"]] = delegated_out["amount"]
                # TODO
                # skip undelegations here, wait for 'return_vesting_delegation'
                # del new_deleg[delegated_out['account']]

        self.delegated_vests_out.append(new_deleg)

    def build(
        self,
        only_ops: Optional[List[str]] = None,
        exclude_ops: Optional[List[str]] = None,
        enable_rewards: bool = False,
        enable_out_votes: bool = False,
        enable_in_votes: bool = False,
    ) -> None:
        """Builds the account history based on all account operations

        :param array only_ops: Limit generator by these
            operations (*optional*)
        :param array exclude_ops: Exclude these operations from
            generator (*optional*)

        """
        if only_ops is None:
            only_ops = []
        if exclude_ops is None:
            exclude_ops = []
        if len(self.timestamps) > 0:
            start_timestamp = self.timestamps[-1]
        else:
            start_timestamp = None
        for op in sorted(self, key=lambda k: k["timestamp"]):
            ts = parse_time(op["timestamp"])
            if start_timestamp is not None:
                # Convert start_timestamp to datetime if it's time or date
                if isinstance(start_timestamp, time):
                    start_timestamp_dt = datetime.combine(datetime.now().date(), start_timestamp)
                elif isinstance(start_timestamp, date):
                    start_timestamp_dt = datetime.combine(start_timestamp, time.min)
                else:
                    start_timestamp_dt = start_timestamp
                if start_timestamp_dt > ts:
                    continue
            if op["type"] in exclude_ops:
                continue
            if len(only_ops) > 0 and op["type"] not in only_ops:
                continue
            self.ops_statistics[op["type"]] += 1
            self.parse_op(
                op,
                only_ops=only_ops,
                enable_rewards=enable_rewards,
                enable_out_votes=enable_out_votes,
                enable_in_votes=enable_in_votes,
            )

    def parse_op(
        self,
        op: Dict[str, Any],
        only_ops: Optional[List[str]] = None,
        enable_rewards: bool = False,
        enable_out_votes: bool = False,
        enable_in_votes: bool = False,
    ) -> None:
        """
        Parse a single account-history operation and update the snapshot's internal state.

        Parses the provided operation dictionary `op` (expects keys like "type" and "timestamp"), converts amounts using the snapshot's blockchain instance, and applies its effect to the snapshot by calling the appropriate update methods (e.g. update, update_rewards, update_out_vote, update_in_vote). Handles Hive-specific operations such as account creation, transfers, vesting/delegation, reward payouts, order fills, conversions, interest, votes, and hardfork-related adjustments. Many other operation types are intentionally ignored.

        Parameters:
            op (dict): A single operation entry from account history. Must contain a parsable "timestamp" and a "type" key; other required keys depend on the operation type.
            only_ops (list): If non-empty, treat operations listed here as affecting balances/votes even when reward/vote-collection flags are disabled.
            enable_rewards (bool): When True, record reward aggregates via update_rewards in addition to applying balance changes.
            enable_out_votes (bool): When True, record outbound votes (this account as voter) via update_out_vote.
            enable_in_votes (bool): When True, record inbound votes (this account as author/recipient) via update_in_vote.

        Returns:
            None
        """
        if only_ops is None:
            only_ops = []
        ts = parse_time(op["timestamp"])

        if op["type"] == "account_create":
            fee_hive = Amount(op["fee"], blockchain_instance=self.blockchain).amount
            fee_vests = self.blockchain.hp_to_vests(
                Amount(op["fee"], blockchain_instance=self.blockchain).amount, timestamp=ts
            )
            if op["new_account_name"] == self.account["name"]:
                self.update(ts, fee_vests, 0, 0)
                return
            if op["creator"] == self.account["name"]:
                self.update(ts, 0, 0, 0, fee_hive * (-1), 0)
                return

        if op["type"] == "account_create_with_delegation":
            fee_hive = Amount(op["fee"], blockchain_instance=self.blockchain).amount
            fee_vests = self.blockchain.hp_to_vests(
                Amount(op["fee"], blockchain_instance=self.blockchain).amount, timestamp=ts
            )
            if op["new_account_name"] == self.account["name"]:
                if Amount(op["delegation"], blockchain_instance=self.blockchain).amount > 0:
                    delegation = {
                        "account": op["creator"],
                        "amount": Amount(op["delegation"], blockchain_instance=self.blockchain),
                    }
                else:
                    delegation = None
                self.update(ts, fee_vests, delegation, 0)
                return

            if op["creator"] == self.account["name"]:
                delegation = {
                    "account": op["new_account_name"],
                    "amount": Amount(op["delegation"], blockchain_instance=self.blockchain),
                }
                self.update(ts, 0, 0, delegation, fee_hive * (-1), 0)
                return

        elif op["type"] == "delegate_vesting_shares":
            vests = Amount(op["vesting_shares"], blockchain_instance=self.blockchain)
            if op["delegator"] == self.account["name"]:
                delegation = {"account": op["delegatee"], "amount": vests}
                self.update(ts, 0, 0, delegation)
                return
            if op["delegatee"] == self.account["name"]:
                delegation = {"account": op["delegator"], "amount": vests}
                self.update(ts, 0, delegation, 0)
                return

        elif op["type"] == "transfer":
            amount = Amount(op["amount"], blockchain_instance=self.blockchain)
            if op["from"] == self.account["name"]:
                if amount.symbol == self.blockchain.blockchain_symbol:
                    self.update(ts, 0, None, None, hive=amount * (-1), hbd=0)
                elif amount.symbol == self.blockchain.backed_token_symbol:
                    self.update(ts, 0, None, None, hive=0, hbd=amount * (-1))
            if op["to"] == self.account["name"]:
                if amount.symbol == self.blockchain.blockchain_symbol:
                    self.update(ts, 0, None, None, hive=amount, hbd=0)
                elif amount.symbol == self.blockchain.backed_token_symbol:
                    self.update(ts, 0, None, None, hive=0, hbd=amount)
            return

        elif op["type"] == "fill_order":
            current_pays = Amount(op["current_pays"], blockchain_instance=self.blockchain)
            open_pays = Amount(op["open_pays"], blockchain_instance=self.blockchain)
            if op["current_owner"] == self.account["name"]:
                if current_pays.symbol == self.blockchain.token_symbol:
                    self.update(ts, 0, None, None, hive=current_pays * (-1), hbd=open_pays)
                elif current_pays.symbol == self.blockchain.backed_token_symbol:
                    self.update(ts, 0, None, None, hive=open_pays, hbd=current_pays * (-1))
            if op["open_owner"] == self.account["name"]:
                if current_pays.symbol == self.blockchain.token_symbol:
                    self.update(ts, 0, None, None, hive=current_pays, hbd=open_pays * (-1))
                elif current_pays.symbol == self.blockchain.backed_token_symbol:
                    self.update(ts, 0, None, None, hive=open_pays * (-1), hbd=current_pays)
            return

        elif op["type"] == "transfer_to_vesting":
            hive_amt = Amount(op["amount"], blockchain_instance=self.blockchain)
            vests = self.blockchain.hp_to_vests(hive_amt.amount, timestamp=ts)
            if op["from"] == self.account["name"] and op["to"] == self.account["name"]:
                self.update(
                    ts, vests, 0, 0, hive_amt * (-1), 0
                )  # power up from and to given account
            elif op["from"] != self.account["name"] and op["to"] == self.account["name"]:
                self.update(ts, vests, 0, 0, 0, 0)  # power up from another account
            else:  # op['from'] == self.account["name"] and op['to'] != self.account["name"]
                self.update(ts, 0, 0, 0, hive_amt * (-1), 0)  # power up to another account
            return

        elif op["type"] == "fill_vesting_withdraw":
            vests = Amount(op["withdrawn"], blockchain_instance=self.blockchain)
            self.update(ts, vests * (-1), None, None, hive=0, hbd=0)
            return

        elif op["type"] == "return_vesting_delegation":
            delegation = {
                "account": None,
                "amount": Amount(op["vesting_shares"], blockchain_instance=self.blockchain),
            }
            self.update(ts, 0, None, delegated_out=delegation)
            return

        elif op["type"] == "claim_reward_balance":
            vests = Amount(op["reward_vests"], blockchain_instance=self.blockchain)
            hive = Amount(op["reward_hive"], blockchain_instance=self.blockchain)
            hbd = Amount(op["reward_hbd"], blockchain_instance=self.blockchain)
            self.update(ts, vests, None, None, hive=hive, hbd=hbd)
            return

        elif op["type"] == "curation_reward":
            if "curation_reward" in only_ops or enable_rewards:
                vests = Amount(op["reward"], blockchain_instance=self.blockchain)
            if "curation_reward" in only_ops:
                self.update(ts, vests, None, None, hive=0, hbd=0)
            if enable_rewards:
                self.update_rewards(ts, vests, 0, 0, 0)
            return

        elif op["type"] == "author_reward":
            if "author_reward" in only_ops or enable_rewards:
                vests = Amount(op["vesting_payout"], blockchain_instance=self.blockchain)
                hive = Amount(op["hive_payout"], blockchain_instance=self.blockchain)
                hbd = Amount(op["hbd_payout"], blockchain_instance=self.blockchain)
            if "author_reward" in only_ops:
                self.update(ts, vests, None, None, hive=hive, hbd=hbd)
            if enable_rewards:
                self.update_rewards(ts, 0, vests, hive, hbd)
            return

        elif op["type"] == "producer_reward":
            vests = Amount(op["vesting_shares"], blockchain_instance=self.blockchain)
            self.update(ts, vests, None, None, hive=0, hbd=0)
            return

        elif op["type"] == "comment_benefactor_reward":
            if op["benefactor"] == self.account["name"]:
                if "reward" in op:
                    vests = Amount(op["reward"], blockchain_instance=self.blockchain)
                    self.update(ts, vests, None, None, hive=0, hbd=0)
                else:
                    vests = Amount(op["vesting_payout"], blockchain_instance=self.blockchain)
                    hive = Amount(op["hive_payout"], blockchain_instance=self.blockchain)
                    hbd = Amount(op["hbd_payout"], blockchain_instance=self.blockchain)
                    self.update(ts, vests, None, None, hive=hive, hbd=hbd)
                return
            else:
                return

        elif op["type"] == "fill_convert_request":
            amount_in = Amount(op["amount_in"], blockchain_instance=self.blockchain)
            amount_out = Amount(op["amount_out"], blockchain_instance=self.blockchain)
            if op["owner"] == self.account["name"]:
                self.update(ts, 0, None, None, hive=amount_out, hbd=amount_in * (-1))
            return

        elif op["type"] == "interest":
            interest = Amount(op["interest"], blockchain_instance=self.blockchain)
            self.update(ts, 0, None, None, hive=0, hbd=interest)
            return

        elif op["type"] == "vote":
            if "vote" in only_ops or enable_out_votes:
                weight = int(op["weight"])
                if op["voter"] == self.account["name"]:
                    self.update_out_vote(ts, weight)
            if "vote" in only_ops or enable_in_votes and op["author"] == self.account["name"]:
                weight = int(op["weight"])
                self.update_in_vote(ts, weight, op)
            return

        elif op["type"] == "hardfork_hive":
            vests = Amount(op["vests_converted"])
            hbd = Amount(op["hbd_transferred"])
            hive = Amount(op["hive_transferred"])
            self.update(ts, vests * (-1), None, None, hive=hive * (-1), hbd=hbd * (-1))

        elif op["type"] in [
            "comment",
            "feed_publish",
            "shutdown_witness",
            "account_witness_vote",
            "witness_update",
            "custom_json",
            "limit_order_create",
            "account_update",
            "account_witness_proxy",
            "limit_order_cancel",
            "comment_options",
            "delete_comment",
            "interest",
            "recover_account",
            "pow",
            "fill_convert_request",
            "convert",
            "request_account_recovery",
            "update_proposal_votes",
        ]:
            return

    def build_sp_arrays(self) -> None:
        """
        Build timelines of own and effective Hive Power (HP) for each stored timestamp.

        For every timestamp in the snapshot, convert the account's own vesting shares and the
        sum of delegated-in/out vesting shares to Hive Power via the blockchain's
        `vests_to_hp` conversion and populate:
        - self.own_sp: HP equivalent of the account's own vesting shares at each timestamp.
        - self.eff_sp: effective HP = own HP + HP delegated in - HP delegated out at each timestamp.

        This method mutates self.own_sp and self.eff_sp in-place and relies on
        self.timestamps, self.own_vests, self.delegated_vests_in, self.delegated_vests_out,
        and self.blockchain.vests_to_hp(timestamp=...).
        """
        self.own_sp = []
        self.eff_sp = []
        for ts, own, din, dout in zip(
            self.timestamps, self.own_vests, self.delegated_vests_in, self.delegated_vests_out
        ):
            sum_in = sum([din[key].amount for key in din])
            sum_out = sum([dout[key].amount for key in dout])
            sp_in = self.blockchain.vests_to_hp(sum_in, timestamp=ts)
            sp_out = self.blockchain.vests_to_hp(sum_out, timestamp=ts)
            sp_own = self.blockchain.vests_to_hp(own, timestamp=ts)

            sp_eff = sp_own + sp_in - sp_out
            self.own_sp.append(sp_own)
            self.eff_sp.append(sp_eff)

    def build_rep_arrays(self) -> None:
        """Build reputation arrays"""
        self.rep_timestamp = [self.timestamps[1]]
        self.rep = [reputation_to_score(0)]
        current_reputation = 0
        for ts, rshares, rep in zip(self.in_vote_timestamp, self.in_vote_rshares, self.in_vote_rep):
            if rep > 0:
                if rshares > 0 or (rshares < 0 and rep > current_reputation):
                    current_reputation += rshares >> 6
            self.rep.append(reputation_to_score(current_reputation))
            self.rep_timestamp.append(ts)

    def build_vp_arrays(self) -> None:
        """
        Build timelines for upvote and downvote voting power.

        Populates the following instance arrays with parallel timestamps and voting-power values:
        - self.vp_timestamp, self.vp: upvoting power timeline
        - self.downvote_vp_timestamp, self.downvote_vp: downvoting power timeline

        The method iterates over recorded outgoing votes (self.out_vote_timestamp / self.out_vote_weight),
        applies Hive vote-regeneration rules (using HIVE_VOTE_REGENERATION_SECONDS and HIVE_100_PERCENT),
        accounts for the HF_21 downvote timing change, and models vote drains via the blockchain's
        _vote/resulting calculation and the account's manabar recharge intervals (account.get_manabar_recharge_timedelta).
        Values are stored as integer percentage units where HIVE_100_PERCENT (typically 10000) represents 100.00%.

        Side effects:
        - Modifies self.vp_timestamp, self.vp, self.downvote_vp_timestamp, and self.downvote_vp in place.
        """
        self.vp_timestamp = [self.timestamps[1]]
        self.vp = [HIVE_100_PERCENT]
        HF_21 = datetime(2019, 8, 27, 15, tzinfo=timezone.utc)
        # Ensure timestamps[1] is datetime for comparison
        ts1 = self.timestamps[1]
        if isinstance(ts1, time):
            ts1 = datetime.combine(datetime.now().date(), ts1)
        elif isinstance(ts1, date):
            ts1 = datetime.combine(ts1, time.min)
        if ts1 is not None and ts1 > HF_21:
            self.downvote_vp_timestamp = [self.timestamps[1]]
        else:
            self.downvote_vp_timestamp = [HF_21]
        self.downvote_vp = [HIVE_100_PERCENT]

        for ts, weight in zip(self.out_vote_timestamp, self.out_vote_weight):
            regenerated_vp = 0
            if ts > HF_21 and weight < 0:
                self.downvote_vp.append(self.downvote_vp[-1])
                if self.downvote_vp[-1] < HIVE_100_PERCENT:
                    regenerated_vp = (
                        ((ts - self.downvote_vp_timestamp[-1]).total_seconds())
                        * HIVE_100_PERCENT
                        / HIVE_VOTE_REGENERATION_SECONDS
                    )
                    self.downvote_vp[-1] += int(regenerated_vp)

                if self.downvote_vp[-1] > HIVE_100_PERCENT:
                    self.downvote_vp[-1] = HIVE_100_PERCENT
                    recharge_time = self.account.get_manabar_recharge_timedelta(
                        {"current_mana_pct": self.downvote_vp[-2] / 100}
                    )
                    # Add full downvote VP once fully charged
                    last_ts = self.downvote_vp_timestamp[-1]
                    if isinstance(last_ts, time):
                        last_ts = datetime.combine(datetime.now().date(), last_ts)
                    elif isinstance(last_ts, date):
                        last_ts = datetime.combine(last_ts, time.min)
                    if last_ts is not None:
                        self.downvote_vp_timestamp.append(last_ts + recharge_time)
                        self.downvote_vp.append(HIVE_100_PERCENT)

                # Add charged downvote VP just before new Vote
                self.downvote_vp_timestamp.append(ts - timedelta(seconds=1))
                self.downvote_vp.append(
                    min([HIVE_100_PERCENT, self.downvote_vp[-1] + regenerated_vp])
                )

                self.downvote_vp[-1] -= (
                    self.blockchain._calc_resulting_vote(HIVE_100_PERCENT, weight) * 4
                )
                # Downvote mana pool is 1/4th of the upvote mana pool, so it gets drained 4 times as quick
                if self.downvote_vp[-1] < 0:
                    # There's most likely a better solution to this that what I did here
                    self.vp.append(self.vp[-1])

                    if self.vp[-1] < HIVE_100_PERCENT:
                        regenerated_vp = (
                            ((ts - self.vp_timestamp[-1]).total_seconds())
                            * HIVE_100_PERCENT
                            / HIVE_VOTE_REGENERATION_SECONDS
                        )
                        self.vp[-1] += int(regenerated_vp)

                    if self.vp[-1] > HIVE_100_PERCENT:
                        self.vp[-1] = HIVE_100_PERCENT
                        recharge_time = self.account.get_manabar_recharge_timedelta(
                            {"current_mana_pct": self.vp[-2] / 100}
                        )
                        # Add full VP once fully charged
                        last_vp_ts = self.vp_timestamp[-1]
                        if isinstance(last_vp_ts, time):
                            last_vp_ts = datetime.combine(datetime.now().date(), last_vp_ts)
                        elif isinstance(last_vp_ts, date):
                            last_vp_ts = datetime.combine(last_vp_ts, time.min)
                        if last_vp_ts is not None:
                            self.vp_timestamp.append(last_vp_ts + recharge_time)
                            self.vp.append(HIVE_100_PERCENT)
                    if self.vp[-1] == HIVE_100_PERCENT and ts - self.vp_timestamp[-1] > timedelta(
                        seconds=1
                    ):
                        # Add charged VP just before new Vote
                        self.vp_timestamp.append(ts - timedelta(seconds=1))
                        self.vp.append(min([HIVE_100_PERCENT, self.vp[-1] + regenerated_vp]))
                    self.vp[-1] += self.downvote_vp[-1] / 4
                    if self.vp[-1] < 0:
                        self.vp[-1] = 0

                    self.vp_timestamp.append(ts)
                    self.downvote_vp[-1] = 0
                self.downvote_vp_timestamp.append(ts)

            else:
                self.vp.append(self.vp[-1])

                if self.vp[-1] < HIVE_100_PERCENT:
                    regenerated_vp = (
                        ((ts - self.vp_timestamp[-1]).total_seconds())
                        * HIVE_100_PERCENT
                        / HIVE_VOTE_REGENERATION_SECONDS
                    )
                    self.vp[-1] += int(regenerated_vp)

                if self.vp[-1] > HIVE_100_PERCENT:
                    self.vp[-1] = HIVE_100_PERCENT
                    recharge_time = self.account.get_manabar_recharge_timedelta(
                        {"current_mana_pct": self.vp[-2] / 100}
                    )
                    # Add full VP once fully charged
                    last_vp_ts = self.vp_timestamp[-1]
                    if isinstance(last_vp_ts, time):
                        last_vp_ts = datetime.combine(datetime.now().date(), last_vp_ts)
                    elif isinstance(last_vp_ts, date):
                        last_vp_ts = datetime.combine(last_vp_ts, time.min)
                    if last_vp_ts is not None:
                        self.vp_timestamp.append(last_vp_ts + recharge_time)
                        self.vp.append(HIVE_100_PERCENT)
                if self.vp[-1] == HIVE_100_PERCENT and ts - self.vp_timestamp[-1] > timedelta(
                    seconds=1
                ):
                    # Add charged VP just before new Vote
                    self.vp_timestamp.append(ts - timedelta(seconds=1))
                    self.vp.append(min([HIVE_100_PERCENT, self.vp[-1] + regenerated_vp]))
                self.vp[-1] -= self.blockchain._calc_resulting_vote(self.vp[-1], weight)
                if self.vp[-1] < 0:
                    self.vp[-1] = 0

                self.vp_timestamp.append(ts)

        if self.account.get_voting_power() == 100:
            self.vp.append(10000)
            recharge_time = self.account.get_manabar_recharge_timedelta(
                {"current_mana_pct": self.vp[-2] / 100}
            )
            last_vp_ts = self.vp_timestamp[-1]
            if isinstance(last_vp_ts, time):
                last_vp_ts = datetime.combine(datetime.now().date(), last_vp_ts)
            elif isinstance(last_vp_ts, date):
                last_vp_ts = datetime.combine(last_vp_ts, time.min)
            if last_vp_ts is not None:
                self.vp_timestamp.append(last_vp_ts + recharge_time)

        if self.account.get_downvoting_power() == 100:
            self.downvote_vp.append(10000)
            recharge_time = self.account.get_manabar_recharge_timedelta(
                {"current_mana_pct": self.downvote_vp[-2] / 100}
            )
            last_downvote_ts = self.downvote_vp_timestamp[-1]
            if isinstance(last_downvote_ts, time):
                last_downvote_ts = datetime.combine(datetime.now().date(), last_downvote_ts)
            elif isinstance(last_downvote_ts, date):
                last_downvote_ts = datetime.combine(last_downvote_ts, time.min)
            if last_downvote_ts is not None:
                self.downvote_vp_timestamp.append(last_downvote_ts + recharge_time)

        self.vp.append(self.account.get_voting_power() * 100)
        self.downvote_vp.append(self.account.get_downvoting_power() * 100)
        self.downvote_vp_timestamp.append(datetime.now(timezone.utc))
        self.vp_timestamp.append(datetime.now(timezone.utc))

    def build_curation_arrays(
        self, end_date: Optional[Union[datetime, date, time]] = None, sum_days: int = 7
    ) -> None:
        """
        Compute curation-per-1000-HP time series and store them in
        self.curation_per_1000_HP_timestamp and self.curation_per_1000_HP.

        The method walks through recorded reward timestamps and curation rewards, converts
        each curation reward (vests) to HP using the blockchain conversion, and divides
        that reward by the effective stake (sp_eff) at the reward time to produce a
        "curation per 1000 HP" value. Values are aggregated into contiguous windows of
        length `sum_days`. Each window's aggregate is appended to
        self.curation_per_1000_HP with the corresponding window end timestamp in
        self.curation_per_1000_HP_timestamp.

        Parameters:
            end_date (datetime.datetime | None): End-boundary for the first aggregation
                window. If None, it is set to the last reward timestamp minus the total
                span of full `sum_days` windows that fit into the reward history.
            sum_days (int): Window length in days for aggregation. Must be > 0.

        Raises:
            ValueError: If sum_days <= 0.

        Notes:
            - Uses self.blockchain.vests_to_hp(vests, timestamp=ts) to convert vests to HP.
            - Uses self.get_data(timestamp=ts, index=index) to obtain the effective stake
              (`sp_eff`) and to advance a cached index for efficient lookups.
            - The per-window aggregation normalizes values to a "per 1000 HP" basis and
              scales them by (7 / sum_days) so the resulting numbers are comparable to a
              7-day baseline.
        """
        self.curation_per_1000_HP_timestamp = []
        self.curation_per_1000_HP = []
        if sum_days <= 0:
            raise ValueError("sum_days must be greater than 0")
        index = 0
        curation_sum = 0
        days = (self.reward_timestamps[-1] - self.reward_timestamps[0]).days // sum_days * sum_days
        if end_date is None:
            end_date = self.reward_timestamps[-1] - timedelta(days=days)
        for ts, vests in zip(self.reward_timestamps, self.curation_rewards):
            if vests == 0:
                continue
            sp = self.blockchain.vests_to_hp(vests, timestamp=ts)
            data = self.get_data(timestamp=ts, index=index)
            index = data["index"]
            if "sp_eff" in data and data["sp_eff"] > 0:
                curation_1k_sp = sp / data["sp_eff"] * 1000 / sum_days * 7
            else:
                curation_1k_sp = 0
            if ts < end_date:
                curation_sum += curation_1k_sp
            else:
                self.curation_per_1000_HP_timestamp.append(end_date)
                self.curation_per_1000_HP.append(curation_sum)
                # Ensure end_date is a datetime for arithmetic
                if isinstance(end_date, datetime):
                    end_date = end_date + timedelta(days=sum_days)
                elif isinstance(end_date, date):
                    end_date = datetime.combine(end_date, time.min, timezone.utc) + timedelta(
                        days=sum_days
                    )
                else:  # time object
                    end_date = datetime.combine(date.today(), end_date, timezone.utc) + timedelta(
                        days=sum_days
                    )
                curation_sum = 0

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return "<{} {}>".format(self.__class__.__name__, str(self.account["name"]))



import json
import unittest

from nectarbase.objects import Amount, Operation


class Testcases(unittest.TestCase):
    def test_Amount(self):
        a = "1.000 HIVE"
        t = Amount(a)
        self.assertEqual(a, t.__str__())
        self.assertEqual(a, str(t))

        t = Amount(a, json_str=True, prefix="STM")
        self.assertEqual(
            {"amount": "1000", "precision": 3, "nai": "@@000000021"}, json.loads(str(t))
        )

        a = {"amount": "3000", "precision": 3, "nai": "@@000000037"}
        t = Amount(a, prefix="STM")
        # self.assertEqual(str(a), t.__str__())
        self.assertEqual(a, json.loads(str(t)))

    def test_Amount_overflow(self):
        a = "0.9999 HIVE"
        t = Amount(a)
        self.assertEqual("0.999 HIVE", t.__str__())
        self.assertEqual("0.999 HIVE", str(t))
        a = "0.9991 HIVE"
        t = Amount(a)
        self.assertEqual("0.999 HIVE", t.__str__())
        self.assertEqual("0.999 HIVE", str(t))

        a = "8.9999 HIVE"
        t = Amount(a)
        self.assertEqual("8.999 HIVE", t.__str__())
        self.assertEqual("8.999 HIVE", str(t))
        a = "8.9991 HIVE"
        t = Amount(a)
        self.assertEqual("8.999 HIVE", t.__str__())
        self.assertEqual("8.999 HIVE", str(t))

        a = "8.19 HIVE"
        t = Amount(a)
        self.assertEqual("8.190 HIVE", t.__str__())
        self.assertEqual("8.190 HIVE", str(t))

        a = "0.0009 HIVE"
        t = Amount(a)
        self.assertEqual("0.000 HIVE", t.__str__())
        self.assertEqual("0.000 HIVE", str(t))

        a = "100.0009 HIVE"
        t = Amount(a)
        self.assertEqual("100.000 HIVE", t.__str__())
        self.assertEqual("100.000 HIVE", str(t))

    def test_Operation(self):
        a = {"amount": "1000", "precision": 3, "nai": "@@000000013"}
        j = ["transfer", {"from": "a", "to": "b", "amount": a, "memo": "c"}]
        o = Operation(j)
        self.assertEqual(o.json()[1], j[1])



from typing import Optional


class Prefix:
    """This class is meant to allow changing the prefix.
    The prefix is used to link a public key to a specific blockchain.
    """

    prefix: str = "STM"

    def set_prefix(self, prefix: Optional[str]) -> None:
        if prefix:
            self.prefix = prefix



from typing import Any, Dict, Union, cast

from .blockchainobject import BlockchainObject
from .exceptions import AssetDoesNotExistsException


class Asset(BlockchainObject):
    """Deals with Assets of the network.

    :param str Asset: Symbol name or object id of an asset
    :param bool lazy: Lazy loading
    :param bool full: Also obtain bitasset-data and dynamic asset dat
    :param Blockchain blockchain_instance: Blockchain instance
    :returns: All data of an asset

    .. note:: This class comes with its own caching function to reduce the
              load on the API server. Instances of this class can be
              refreshed with ``Asset.refresh()``.
    """

    type_id = 3

    def __init__(
        self,
        asset: Union[str, int],
        lazy: bool = False,
        full: bool = False,
        blockchain_instance: Any = None,
        **kwargs,
    ) -> None:
        self.full = full
        super().__init__(
            asset, lazy=lazy, full=full, blockchain_instance=blockchain_instance, **kwargs
        )
        # self.refresh()

    def refresh(self) -> None:
        """Refresh the data from the API server"""
        self.chain_params = self.blockchain.get_network()
        if self.chain_params is None:
            from nectargraphenebase.chains import known_chains

            self.chain_params = known_chains["HIVE"]
        cast(Dict[str, Any], self)["asset"] = ""
        found_asset = False

        # Store original identifier before it gets overwritten
        original_identifier = self.identifier
        if hasattr(original_identifier, "symbol"):
            # If identifier is an Asset object, get its symbol
            original_identifier = original_identifier.symbol
        elif hasattr(original_identifier, "identifier"):
            # If identifier has an identifier attribute, get its string representation
            original_identifier = str(original_identifier)
        elif not isinstance(original_identifier, (str, int)):
            # Convert to string if it's not already a string or int
            original_identifier = str(original_identifier)

        for asset in self.chain_params["chain_assets"]:
            if original_identifier in [asset["symbol"], str(asset["asset"]), str(asset["id"])]:
                cast(Dict[str, Any], self)["asset"] = asset["asset"]
                cast(Dict[str, Any], self)["precision"] = asset["precision"]
                cast(Dict[str, Any], self)["id"] = asset["id"]
                cast(Dict[str, Any], self)["symbol"] = asset["symbol"]
                found_asset = True
                break
        if not found_asset:
            raise AssetDoesNotExistsException(
                f"{original_identifier} chain_assets:{self.chain_params['chain_assets']}"
            )

    @property
    def symbol(self) -> str:
        return cast(Dict[str, Any], self)["symbol"]

    @property
    def asset(self) -> str:
        return cast(Dict[str, Any], self)["asset"]

    @property
    def precision(self) -> int:
        return cast(Dict[str, Any], self)["precision"]

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Asset):
            return (
                cast(Dict[str, Any], self)["symbol"] == other["symbol"]
                and cast(Dict[str, Any], self)["asset"] == other["asset"]
                and cast(Dict[str, Any], self)["precision"] == other["precision"]
            )
        if isinstance(other, dict):
            symbol = other["symbol"] if "symbol" in other else None
            asset = other["asset"] if "asset" in other else None
            precision = other["precision"] if "precision" in other else None
            return (
                cast(Dict[str, Any], self)["symbol"] == symbol
                and cast(Dict[str, Any], self)["asset"] == asset
                and cast(Dict[str, Any], self)["precision"] == precision
            )
        if isinstance(other, (str, int)):
            return cast(Dict[str, Any], self)["symbol"] == other
        return False

    def __ne__(self, other: object) -> bool:
        if isinstance(other, Asset):
            return (
                cast(Dict[str, Any], self)["symbol"] != other["symbol"]
                or cast(Dict[str, Any], self)["asset"] != other["asset"]
                or cast(Dict[str, Any], self)["precision"] != other["precision"]
            )
        if isinstance(other, dict):
            symbol = other["symbol"] if "symbol" in other else None
            asset = other["asset"] if "asset" in other else None
            precision = other["precision"] if "precision" in other else None
            return (
                cast(Dict[str, Any], self)["symbol"] != symbol
                or cast(Dict[str, Any], self)["asset"] != asset
                or cast(Dict[str, Any], self)["precision"] != precision
            )
        if isinstance(other, (str, int)):
            return cast(Dict[str, Any], self)["symbol"] != other
        return True



#: Object types for object ids
from typing import Dict

object_type: Dict[str, int] = {}
object_type["null"] = 0
object_type["base"] = 1
object_type["account"] = 2
object_type["OBJECT_TYPE_COUNT"] = 3



import logging
from datetime import timedelta

from nectar import exceptions
from nectar.account import Account
from nectar.blockchain import Blockchain
from nectar.comment import Comment
from nectar.utils import construct_authorperm

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class WatchingTheWatchers:
    def __init__(self):
        self.dvcount = 0
        self.account_type = dict()
        self.account_relations = dict()
        self.by_voter = dict()
        self.by_target = dict()
        self.by_pair = dict()

    def update(self, downvoter, downvoted, dvpower, flagpower):
        pair = downvoter + "/" + downvoted
        if downvoter not in self.by_voter:
            self.by_voter[downvoter] = [0.0, 0.0]
        if downvoted not in self.by_target:
            self.by_target[downvoted] = [0.0, 0.0]
        if pair not in self.by_pair:
            self.by_pair[pair] = [0.0, 0.0]
        self.by_voter[downvoter][0] = self.by_voter[downvoter][0] + dvpower
        self.by_voter[downvoter][1] = self.by_voter[downvoter][1] + flagpower
        self.by_target[downvoted][0] = self.by_target[downvoted][0] + dvpower
        self.by_target[downvoted][1] = self.by_target[downvoted][1] + flagpower
        self.by_pair[pair][0] = self.by_pair[pair][0] + dvpower
        self.by_pair[pair][1] = self.by_pair[pair][1] + flagpower
        self.dvcount = self.dvcount + 1
        if self.dvcount % 100 == 0:
            print(self.dvcount, "downvotes so far.")

    def set_account_info(self, account, fish, related):
        self.account_type[account] = fish
        if len(related) > 0:
            self.account_relations[account] = related

    def report(self):
        print("[REPORT]")
        print(" * account_type :", self.account_type)
        print()
        print(" * account_relations :", self.account_relations)
        print()
        print(" * by voter :", self.by_voter)
        print()
        print(" * by target :", self.by_target)
        print()
        print(" * by pair :", self.by_pair)
        print()
        self.dvcount = 0
        self.account_type = dict()
        self.account_relations = dict()
        self.by_voter = dict()
        self.by_target = dict()
        self.by_pair = dict()


class WatchingTheWatchersBot:
    def __init__(self, wtw):
        self.stopped = None
        self.wtw = wtw
        self.looked_up = set()

    def vote(self, vote_event):
        """
        Process a vote event and record downvote statistics and account metadata.

        When the incoming vote_event represents a negative vote (weight < 0), this method:
        - Extracts the corresponding Comment and inspects its active_votes to find the voter's negative vote(s), splitting any contribution into `downvote_power` and `flag_power`, then records those values with the WatchingTheWatchers instance.
        - Ensures voter and author account information is loaded once: looks up account objects, computes vesting power (VP) from vesting shares, received vesting shares, and delegated vesting shares, then classifies accounts into fish categories ("redfish", "minnow", "dolphin", "orca", "whale") by VP thresholds.
        - Collects related accounts from the account's `recovery_account` (excluded when empty or equal to "hive") and `proxy`, and stores that metadata.

        Parameters:
            vote_event (dict): A vote operation dictionary with at least the keys "weight", "voter", "author", and "permlink". For negative weights this method reads the referenced comment's "active_votes" to compute downvote and flag powers.

        Returns:
            None
        """

        def process_vote_content(event):
            start_rshares = 0.0
            for vote in event["active_votes"]:
                if vote["voter"] == vote_event["voter"] and float(vote["rshares"]) < 0:
                    if start_rshares + float(vote["rshares"]) < 0:
                        flag_power = 0 - start_rshares - float(vote["rshares"])
                    else:
                        flag_power = 0
                    downvote_power = 0 - vote["rshares"] - flag_power
                    self.wtw.update(vote["voter"], vote_event["author"], downvote_power, flag_power)

        def lookup_accounts(acclist):
            """
            Lookup account records for the given account names, compute each account's vesting power (VP) and a fish-category based on VP, collect related accounts (recovery_account except "hive" and non-empty proxy), and store the results by calling self.wtw.set_account_info(account, fish, related). If any related accounts (recovery or proxy) are found and not already processed, they are queued for recursive lookup.

            Parameters:
                acclist (list[str]): List of account names to fetch and process.

            Notes:
                - VP is computed as (vesting_shares + received_vesting_shares - delegated_vesting_shares) / 1_000_000.
                - Fish categories: "redfish" (default), "minnow" (VP >= 1), "dolphin" (VP >= 10), "orca" (VP >= 100), "whale" (VP > 1000).
                - Prints an "OOPS" line if the number of fetched Account objects does not match the input list length.
                - Side effects: calls self.wtw.set_account_info and may call lookup_accounts recursively for related accounts not yet seen.
            """

            def user_info(accounts):
                if len(acclist) != len(accounts):
                    print("OOPS:", len(acclist), len(accounts), acclist)
                for index in range(0, len(accounts)):
                    a = accounts[index]
                    account = acclist[index]
                    vp = (
                        a["vesting_shares"].amount
                        + a["received_vesting_shares"].amount
                        - a["delegated_vesting_shares"].amount
                    ) / 1000000.0
                    fish = "redfish"
                    if vp >= 1.0:
                        fish = "minnow"
                    if vp >= 10.0:
                        fish = "dolphin"
                    if vp >= 100:
                        fish = "orca"
                    if vp > 1000:
                        fish = "whale"
                    racc = None
                    proxy = None
                    related = list()
                    if a["recovery_account"] != "hive" and a["recovery_account"] != "":
                        related.append(a["recovery_account"])
                    if a["proxy"] != "":
                        related.append(a["proxy"])
                    self.wtw.set_account_info(account, fish, related)
                    accl2 = list()
                    if racc is not None and racc not in self.looked_up:
                        accl2.append(racc)
                    if proxy is not None and proxy not in self.looked_up:
                        accl2.append(proxy)
                    if len(accl2) > 0:
                        lookup_accounts(accl2)

            accounts = []
            for a in acclist:
                accounts.append(Account(a))
            user_info(accounts)

        if vote_event["weight"] < 0:
            authorperm = construct_authorperm(vote_event["author"], vote_event["permlink"])
            # print(authorperm)
            try:
                process_vote_content(Comment(authorperm))
            except exceptions.ContentDoesNotExistsException:
                print("Could not find Comment: %s" % (authorperm))
            al = list()
            if vote_event["voter"] not in self.looked_up:
                al.append(vote_event["voter"])
                self.looked_up.add(vote_event["voter"])
            if vote_event["author"] not in self.looked_up:
                al.append(vote_event["author"])
                self.looked_up.add(vote_event["author"])
            if len(al) > 0:
                lookup_accounts(al)


if __name__ == "__main__":
    wtw = WatchingTheWatchers()
    tb = WatchingTheWatchersBot(wtw)
    blockchain = Blockchain()
    threading = True
    thread_num = 16
    cur_block = blockchain.get_current_block()
    stop = cur_block.identifier
    startdate = cur_block.time() - timedelta(days=1)
    start = blockchain.get_estimated_block_num(startdate, accurate=True)
    for vote in blockchain.stream(
        opNames=["vote"], start=start, stop=stop, threading=threading, thread_num=thread_num
    ):
        tb.vote(vote)
    wtw.report()



import hashlib
import json
import logging
from binascii import hexlify, unhexlify
from collections import OrderedDict
from typing import Any, Dict, List, Optional, Union

import ecdsa
from asn1crypto.core import OctetString

from .bip32 import parse_path
from .chains import known_chains
from .objects import Operation, isArgsThisClass
from .types import (
    Array,
    JsonObj,
    PointInTime,
    Set,
    Signature,
    String,
    Uint16,
    Uint32,
    Varint32,
)
from .types import (
    Optional as GrapheneOptional,
)

log = logging.getLogger(__name__)


class GrapheneObjectASN1:
    """Core abstraction class

    This class is used for any JSON reflected object in Graphene.

    * ``instance.__json__()``: encodes data into json format
    * ``bytes(instance)``: encodes data into wire format
    * ``str(instances)``: dumps json object as string

    """

    def __init__(self, data: Any = None) -> None:
        self.data = data

    def __bytes__(self) -> bytes:
        if self.data is None:
            return b""
        b = b""
        output = b""
        for name, value in list(self.data.items()):
            if name == "operations":
                for operation in value:
                    if isinstance(value, str):
                        b = bytes(operation, "utf-8")
                    else:
                        b = bytes(operation)
                    output += OctetString(b).dump()
            elif name != "signatures":
                if isinstance(value, str):
                    b = bytes(value, "utf-8")
                else:
                    b = bytes(value)
                output += OctetString(b).dump()
        return output

    def __json__(self) -> Dict[str, Any]:
        if self.data is None:
            return {}
        d = {}  # JSON output is *not* ordered
        for name, value in list(self.data.items()):
            if isinstance(value, GrapheneOptional) and value.isempty():
                continue

            if isinstance(value, String):
                d.update({name: str(value)})
            else:
                try:
                    d.update({name: JsonObj(value)})
                except Exception:
                    d.update({name: value.__str__()})
        return d

    def __str__(self) -> str:
        return json.dumps(self.__json__())

    def toJson(self) -> Dict[str, Any]:
        return self.__json__()

    def json(self) -> Dict[str, Any]:
        return self.__json__()


class Unsigned_Transaction(GrapheneObjectASN1):
    """Create an unsigned transaction with ASN1 encoder for using it with ledger

    :param num ref_block_num:
    :param num ref_block_prefix:
    :param str expiration: expiration date
    :param array operations:  array of operations
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if isArgsThisClass(self, args):
            self.data = args[0].data
        else:
            if len(args) == 1 and len(kwargs) == 0:
                kwargs = args[0]
            kwargs.pop("prefix", "STM")
            if "extensions" not in kwargs:
                kwargs["extensions"] = Set([])
            elif not kwargs.get("extensions"):
                kwargs["extensions"] = Set([])
            if "signatures" not in kwargs:
                kwargs["signatures"] = Array([])
            else:
                kwargs["signatures"] = Array(
                    [Signature(unhexlify(a)) for a in kwargs["signatures"]]
                )
            operations_count = 0
            if "operations" in kwargs:
                operations_count = len(kwargs["operations"])
                # opklass = self.getOperationKlass()
                # if all([not isinstance(a, opklass) for a in kwargs["operations"]]):
                #    kwargs['operations'] = Array([opklass(a, ) for a in kwargs["operations"]])
                # else:
                #    kwargs['operations'] = (kwargs["operations"])

            super().__init__(
                OrderedDict(
                    [
                        ("ref_block_num", Uint16(kwargs["ref_block_num"])),
                        ("ref_block_prefix", Uint32(kwargs["ref_block_prefix"])),
                        ("expiration", PointInTime(kwargs["expiration"])),
                        ("operations_count", Varint32(operations_count)),
                        ("operations", kwargs["operations"]),
                        ("extensions", kwargs["extensions"]),
                        ("signatures", kwargs["signatures"]),
                    ]
                )
            )

    @property
    def id(self) -> str:
        """The transaction id of this transaction"""
        # Store signatures temporarily since they are not part of
        # transaction id
        sigs = self.data["signatures"]
        self.data.pop("signatures", None)

        # Generage Hash of the seriliazed version
        h = hashlib.sha256(bytes(self)).digest()

        # recover signatures
        self.data["signatures"] = sigs

        # Return properly truncated tx hash
        return hexlify(h[:20]).decode("ascii")

    def getOperationKlass(self) -> type[Operation]:
        return Operation

    def derSigToHexSig(self, s: bytes) -> str:
        """Format DER to HEX signature"""
        s, junk = ecdsa.der.remove_sequence(unhexlify(s))
        if junk:
            log.debug("JUNK: %s", hexlify(junk).decode("ascii"))
        if not (junk == b""):
            raise AssertionError()
        x, s = ecdsa.der.remove_integer(s)
        y, s = ecdsa.der.remove_integer(s)
        return "{:064x}{:064x}".format(x, y)

    def getKnownChains(self) -> Dict[str, Any]:
        return known_chains

    def getChainParams(self, chain: Union[str, Dict[str, Any]]) -> Dict[str, Any]:
        # Which network are we on:
        chains = self.getKnownChains()
        if isinstance(chain, str) and chain in chains:
            chain_params = chains[chain]
        elif isinstance(chain, dict):
            chain_params = chain
        else:
            raise Exception("sign() only takes a string or a dict as chain!")
        if "chain_id" not in chain_params:
            raise Exception("sign() needs a 'chain_id' in chain params!")
        return chain_params

    def deriveDigest(self, chain: Union[str, Dict[str, Any]]) -> None:
        chain_params = self.getChainParams(chain)
        # Chain ID
        self.chainid = chain_params["chain_id"]

        # Do not serialize signatures
        sigs = self.data["signatures"]
        self.data["signatures"] = []

        # Get message to sign
        #   bytes(self) will give the wire formated data according to
        #   GrapheneObject and the data given in __init__()
        self.message = OctetString(unhexlify(self.chainid)).dump()
        for name, value in list(self.data.items()):
            if name == "operations":
                for operation in value:
                    if isinstance(value, str):
                        b = bytes(operation, "utf-8")
                    else:
                        b = bytes(operation)
                    self.message += OctetString(b).dump()
            elif name != "signatures":
                if isinstance(value, str):
                    b = bytes(value, "utf-8")
                else:
                    b = bytes(value)
                self.message += OctetString(b).dump()

        self.digest = hashlib.sha256(self.message).digest()

        # restore signatures
        self.data["signatures"] = sigs

    def build_path(self, role: str, account_index: int, key_index: int) -> str:
        if role == "owner":
            return "48'/13'/0'/%d'/%d'" % (account_index, key_index)
        elif role == "active":
            return "48'/13'/1'/%d'/%d'" % (account_index, key_index)
        elif role == "posting":
            return "48'/13'/4'/%d'/%d'" % (account_index, key_index)
        elif role == "memo":
            return "48'/13'/3'/%d'/%d'" % (account_index, key_index)
        else:
            raise ValueError(f"Unknown role: {role}")

    def build_apdu(
        self, path: str = "48'/13'/0'/0'/0'", chain: Optional[Union[str, Dict[str, Any]]] = None
    ) -> List[bytes]:
        if chain is None:
            raise ValueError("chain parameter is required for build_apdu")
        self.deriveDigest(chain)
        parsed_path = parse_path(path, as_bytes=True)
        path_bytes = unhexlify(bytes(parsed_path))

        message = self.message
        path_size = int(len(path_bytes) / 4)
        message_size = len(message)

        offset = 0
        first = True
        result = []
        while offset < message_size:
            chunk = message[offset : offset + 255]
            total_size = len(chunk)
            if first:
                apdu = (
                    unhexlify("d4040000")
                    + bytes([total_size])
                    + bytes([path_size])
                    + path_bytes
                    + chunk
                )
                first = False
            else:
                total_size = len(chunk)
                apdu = unhexlify("d4048000") + bytes([total_size]) + chunk
            result.append(apdu)
            offset += len(chunk)
        return result

    def build_apdu_pubkey(
        self, path: str = "48'/13'/0'/0'/0'", request_screen_approval: bool = False
    ) -> bytes:
        parsed_path = parse_path(path, as_bytes=True)
        path_bytes = unhexlify(bytes(parsed_path))
        if not request_screen_approval:
            return (
                unhexlify("d4020001")
                + bytes([int(len(path_bytes)) + 1])
                + bytes([int(len(path_bytes) / 4)])
                + path_bytes
            )
        else:
            return (
                unhexlify("d4020101")
                + bytes([int(len(path_bytes)) + 1])
                + bytes([int(len(path_bytes) / 4)])
                + path_bytes
            )



from __future__ import annotations

import json
import warnings
from datetime import date, datetime, timezone
from typing import Any

from prettytable import PrettyTable

from nectarapi.exceptions import NoMethodWithName, RPCError
from nectarbase import operations

from .account import Account
from .amount import Amount
from .blockchainobject import BlockchainObject
from .exceptions import WitnessDoesNotExistsException
from .instance import shared_blockchain_instance
from .utils import formatTimeString, parse_time


class Witness(BlockchainObject):
    """Read data about a witness in the chain

    :param str owner: Name of the witness
    :param bool lazy: Use lazy loading
    :param bool full: Get full data about witness
    :param nectar.nectar.nectar blockchain_instance: nectar
        instance to use when accessing the RPC
    """

    type_id = 3

    def __init__(
        self,
        owner: str | dict[str, Any],
        full: bool = False,
        lazy: bool = False,
        blockchain_instance: Any | None = None,
        **kwargs: Any,
    ) -> None:
        # Warn about any unused kwargs to maintain backward compatibility
        """
        Initialize a Witness object representing a blockchain witness.

        Parameters:
            owner (str | dict): Witness owner account name or a dictionary of witness fields. If a dict is provided, it will be parsed into the internal witness representation.
            full (bool): If True, load full witness data when available; otherwise keep a lighter representation.
            lazy (bool): If True, defer network loading until data is accessed.

        Notes:
            - `blockchain_instance` defaults to the shared blockchain instance when not provided.
            - Any unexpected keyword arguments are accepted for backward compatibility but will trigger a DeprecationWarning and be ignored.
        """
        if kwargs:
            for key in kwargs:
                warnings.warn(
                    f"Unexpected keyword argument '{key}' passed to Witness.__init__. "
                    "This may be a deprecated parameter and will be ignored.",
                    DeprecationWarning,
                    stacklevel=2,
                )
        self.full = full
        self.lazy = lazy
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if isinstance(owner, dict):
            owner = self._parse_json_data(owner)
        super().__init__(
            owner, lazy=lazy, full=full, id_item="owner", blockchain_instance=self.blockchain
        )

    def refresh(self) -> None:
        """
        Refresh the witness data from the blockchain and reinitialize this object.

        If the witness identifier is empty or the blockchain is not connected, the method returns early.
        Fetches witness data via the configured RPC, parses
        timestamps and numeric fields via _parse_json_data, and reinitializes the Witness instance with the
        retrieved data (respecting this object's lazy/full flags and blockchain instance).

        Raises:
            WitnessDoesNotExistsException: If no witness information is found for the current identifier.
        """
        if not self.identifier:
            return
        if not self.blockchain.is_connected():
            return
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        witness = self.blockchain.rpc.find_witnesses({"owners": [self.identifier]})["witnesses"]
        if len(witness) > 0:
            witness = witness[0]
        if not witness:
            raise WitnessDoesNotExistsException(self.identifier)
        witness = self._parse_json_data(witness)
        super().__init__(
            witness,
            id_item="owner",
            lazy=self.lazy,
            full=self.full,
            blockchain_instance=self.blockchain,
        )

    def _parse_json_data(self, witness: dict) -> dict:
        parse_times = [
            "created",
            "last_sbd_exchange_update",
            "hardfork_time_vote",
            "last_hbd_exchange_update",
        ]
        for p in parse_times:
            if p in witness and isinstance(witness.get(p), str):
                witness[p] = parse_time(witness.get(p, "1970-01-01T00:00:00"))
        parse_int = [
            "votes",
            "virtual_last_update",
            "virtual_position",
            "virtual_scheduled_time",
        ]
        for p in parse_int:
            if p in witness and isinstance(witness.get(p), str):
                witness[p] = int(witness.get(p, "0"))
        return witness

    def json(self) -> dict[str, Any]:
        output = self.copy()
        parse_times = [
            "created",
            "last_sbd_exchange_update",
            "hardfork_time_vote",
            "last_hbd_exchange_update",
        ]
        for p in parse_times:
            if p in output:
                p_date = output.get(p, datetime(1970, 1, 1, 0, 0))
                if isinstance(p_date, (datetime, date)):
                    output[p] = formatTimeString(p_date)
                else:
                    output[p] = p_date
        parse_int = [
            "votes",
            "virtual_last_update",
            "virtual_position",
            "virtual_scheduled_time",
        ]
        for p in parse_int:
            if p in output and isinstance(output[p], int):
                output[p] = str(output[p])
        return json.loads(str(json.dumps(output)))

    @property
    def account(self) -> Account:
        return Account(self["owner"], blockchain_instance=self.blockchain)

    @property
    def is_active(self) -> bool:
        return (
            len(self["signing_key"]) > 3
            and self["signing_key"][3:] != "1111111111111111111111111111111114T1Anm"
        )

    def feed_publish(self, base, quote=None, account=None):
        """
        Publish a witness feed price (exchange rate) to the blockchain.

        Accepts the base and quote as Amount objects, strings, or numeric values and submits a Feed_publish operation using the provided account (defaults to the witness owner).

        Parameters:
            base: Amount | str | number
                The base side of the exchange_rate (must use the blockchain's backed token symbol).
            quote: Amount | str | number, optional
                The quote side of the exchange_rate. Defaults to "1.000 <TOKEN>" where <TOKEN> is the blockchain token_symbol.
            account: str | Account, optional
                Account name or Account object used to sign and publish the feed. If omitted, the witness owner is used.

        Returns:
            The result returned by blockchain.finalizeOp (typically the broadcast/transaction result).

        Raises:
            ValueError: If no account is provided and the witness has no owner.
            AssertionError: If the resolved base or quote symbols do not match the blockchain's expected backed_token_symbol and token_symbol, respectively.
        """
        quote = quote if quote is not None else "1.000 %s" % (self.blockchain.token_symbol)
        if not account:
            account = self["owner"]
        if not account:
            raise ValueError("You need to provide an account")

        account = Account(account, blockchain_instance=self.blockchain)
        if isinstance(base, Amount):
            base = Amount(base, blockchain_instance=self.blockchain)
        elif isinstance(base, str):
            base = Amount(base, blockchain_instance=self.blockchain)
        else:
            base = Amount(
                base, self.blockchain.backed_token_symbol, blockchain_instance=self.blockchain
            )

        if isinstance(quote, Amount):
            quote = Amount(quote, blockchain_instance=self.blockchain)
        elif isinstance(quote, str):
            quote = Amount(quote, blockchain_instance=self.blockchain)
        else:
            quote = Amount(quote, self.blockchain.token_symbol, blockchain_instance=self.blockchain)

        if not base.symbol == self.blockchain.backed_token_symbol:
            raise AssertionError()
        if not quote.symbol == self.blockchain.token_symbol:
            raise AssertionError()
        op = operations.Feed_publish(
            **{
                "publisher": account["name"],
                "exchange_rate": {
                    "base": base,
                    "quote": quote,
                },
                "prefix": self.blockchain.prefix,
                "json_str": True,
            }
        )
        return self.blockchain.finalizeOp(op, account, "active")

    def update_witness(self, signing_key, url, props, account=None):
        """Update witness

        :param str signing_key: Signing key
        :param str url: URL
        :param dict props: Properties
        :param str account: (optional) witness account name

        Properties:::

            {
                "account_creation_fee": x,
                "maximum_block_size": x,
                "sbd_interest_rate": x,
            }

        """
        if not account:
            account = self["owner"]
        return self.blockchain.witness_update(signing_key, url, props, account=account)

    def update(self, *args: Any, **kwargs: Any) -> Any:
        """Compatibility shim to avoid clashing with MutableMapping.update"""
        return self.update_witness(*args, **kwargs)

    @staticmethod
    def get_witness_by_account(
        account_name: str, blockchain_instance: Any | None = None
    ) -> dict[str, Any] | None:
        """
        Fetch witness information by account name using the `get_witness_by_account` RPC call.

        :param str account_name: Name of the witness account
        :param blockchain_instance: Nectar instance (optional)
        :return: Witness data dictionary or None if not found/error
        """
        blockchain = blockchain_instance or shared_blockchain_instance()
        try:
            return blockchain.rpc.get_witness_by_account(account_name)
        except (RPCError, NoMethodWithName):
            return None


class WitnessesObject(list):
    def printAsTable(self, sort_key="votes", reverse=True, return_str=False, **kwargs):
        no_feed = False
        if (
            len(self) > 0
            and "sbd_exchange_rate" not in self[0]
            and "hbd_exchange_rate" not in self[0]
        ):
            table_header = ["Name", "Votes [PV]", "Disabled", "Missed", "Fee", "Size", "Version"]
            no_feed = True
        else:
            table_header = [
                "Name",
                "Votes [PV]",
                "Disabled",
                "Missed",
                "Feed base",
                "Feed quote",
                "Feed update",
                "Fee",
                "Size",
                "Interest",
                "Version",
            ]
        if "sbd_exchange_rate" in self[0]:
            bd_exchange_rate = "sbd_exchange_rate"
            bd_interest_rate = "sbd_interest_rate"
            last_bd_exchange_update = "last_sbd_exchange_update"
        else:
            bd_exchange_rate = "hbd_exchange_rate"
            bd_interest_rate = "hbd_interest_rate"
            last_bd_exchange_update = "last_hbd_exchange_update"
        t = PrettyTable(table_header)
        t.align = "l"
        if sort_key == "base":
            sortedList = sorted(
                self, key=lambda self: self[bd_exchange_rate]["base"], reverse=reverse
            )
        elif sort_key == "quote":
            sortedList = sorted(
                self, key=lambda self: self[bd_exchange_rate]["quote"], reverse=reverse
            )
        elif sort_key == "last_sbd_exchange_update" or sort_key == "last_hbd_exchange_update":
            sortedList = sorted(
                self,
                key=lambda self: (
                    datetime.now(timezone.utc) - self[last_bd_exchange_update]
                ).total_seconds(),
                reverse=reverse,
            )
        elif sort_key == "account_creation_fee":
            sortedList = sorted(
                self, key=lambda self: self["props"]["account_creation_fee"], reverse=reverse
            )
        elif sort_key == "sbd_interest_rate" or sort_key == "hbd_interest_rate":
            sortedList = sorted(
                self, key=lambda self: self["props"][bd_interest_rate], reverse=reverse
            )
        elif sort_key == "maximum_block_size":
            sortedList = sorted(
                self, key=lambda self: self["props"]["maximum_block_size"], reverse=reverse
            )
        elif sort_key == "votes":
            sortedList = sorted(self, key=lambda self: int(self[sort_key]), reverse=reverse)
        else:
            sortedList = sorted(self, key=lambda self: self[sort_key], reverse=reverse)
        for witness in sortedList:
            disabled = ""
            if not witness.is_active:
                disabled = "yes"

            if no_feed:
                t.add_row(
                    [
                        witness["owner"],
                        str(round(int(witness["votes"]) / 1e15, 2)),
                        disabled,
                        str(witness["total_missed"]),
                        str(witness["props"]["account_creation_fee"]),
                        str(witness["props"]["maximum_block_size"]),
                        witness["running_version"],
                    ]
                )
            else:
                # Handle both string and datetime formats for exchange update timestamps
                exchange_update = witness[last_bd_exchange_update]
                if isinstance(exchange_update, str):
                    exchange_update_dt = datetime.strptime(
                        exchange_update, "%Y-%m-%dT%H:%M:%S"
                    ).replace(tzinfo=timezone.utc)
                else:
                    exchange_update_dt = exchange_update
                td = datetime.now(timezone.utc) - exchange_update_dt
                t.add_row(
                    [
                        witness["owner"],
                        str(round(int(witness["votes"]) / 1e15, 2)),
                        disabled,
                        str(witness["total_missed"]),
                        str(
                            Amount(
                                witness[bd_exchange_rate]["base"],
                                blockchain_instance=witness.blockchain,
                            )
                        ),
                        str(
                            Amount(
                                witness[bd_exchange_rate]["quote"],
                                blockchain_instance=witness.blockchain,
                            )
                        ),
                        str(td.days)
                        + " days "
                        + str(td.seconds // 3600)
                        + ":"
                        + str((td.seconds // 60) % 60),
                        str(
                            Amount(
                                witness["props"]["account_creation_fee"],
                                blockchain_instance=witness.blockchain,
                            )
                        ),
                        str(witness["props"]["maximum_block_size"]),
                        str(witness["props"][bd_interest_rate] / 100) + " %",
                        witness["running_version"],
                    ]
                )
        if return_str:
            return t.get_string(**kwargs)
        else:
            print(t.get_string(**kwargs))

    def get_votes_sum(self):
        vote_sum = 0
        for witness in self:
            vote_sum += int(witness["votes"])
        return vote_sum

    def __contains__(self, item: object, /) -> bool:  # type: ignore[override]
        from .account import Account

        if isinstance(item, Account):
            name = item["name"]
        elif isinstance(item, str):
            name = item
        elif len(self) > 0 and hasattr(self[0], "blockchain") and isinstance(item, dict):
            account = Account(item, blockchain_instance=self[0].blockchain)
            name = account["name"]
        else:
            # Fallback to string comparison
            name = str(item)

        return any([name == x["owner"] for x in self])

    def __str__(self):
        return self.printAsTable(return_str=True)

    def __repr__(self):
        return "<{} {}>".format(
            self.__class__.__name__,
            str(getattr(self, "identifier", f"list_{len(self)}")),
        )


class GetWitnesses(WitnessesObject):
    """Obtain a list of witnesses

    :param list name_list: list of witneses to fetch
    :param int batch_limit: (optional) maximum number of witnesses
        to fetch per call, defaults to 100
    :param nectar.nectar.nectar blockchain_instance: nectar instance to use when
        accessing the RPC

    .. code-block:: python

        from nectar.witness import GetWitnesses
        w = GetWitnesses(["gtg", "jesta"])
        print(w[0].json())
        print(w[1].json())

    """

    def __init__(
        self,
        name_list,
        batch_limit=100,
        lazy=False,
        full=True,
        blockchain_instance=None,
    ):
        """
        Initialize the GetWitnesses collection by fetching witness objects for the given list of account names.

        Names are fetched in batches (size controlled by `batch_limit`). If no blockchain connection is available the initializer returns early and the collection remains empty.

        Parameters:
            name_list (Iterable[str]): Account names of witnesses to retrieve.
            batch_limit (int): Maximum number of names to request per batch.
            lazy (bool): If True, create Witness objects in lazy-loading mode.
            full (bool): If True, create Witness objects with full data loaded.
        """
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        if not self.blockchain.is_connected():
            return
        witnesses = []
        name_cnt = 0
        while name_cnt < len(name_list):
            self.blockchain.rpc.set_next_node_on_empty_reply(False)
            witnesses += self.blockchain.rpc.find_witnesses(
                {"owners": name_list[name_cnt : batch_limit + name_cnt]}
            )["witnesses"]
            name_cnt += batch_limit
        self.identifier = ""
        super().__init__(
            [
                Witness(x, lazy=lazy, full=full, blockchain_instance=self.blockchain)
                for x in witnesses
            ]
        )


class Witnesses(WitnessesObject):
    """Obtain a list of **active** witnesses and the current schedule

    :param nectar.nectar.nectar blockchain_instance: nectar instance to use when
        accessing the RPC

    .. code-block:: python

       >>> from nectar.witness import Witnesses
       >>> Witnesses()
       <Witnesses >

    """

    def __init__(self, lazy=False, full=True, blockchain_instance=None):
        """
        Initialize a Witnesses collection and load active witnesses from the configured blockchain.

        Parameters:
            lazy (bool): If True, create Witness objects without fetching full data (deferred loading).
            full (bool): If True, eager-load full witness data when constructing each Witness.

        Notes:
            Resolves the blockchain instance from `blockchain_instance` or the shared default and immediately calls `refresh()` to populate the list of active witnesses.
        """
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.lazy = lazy
        self.full = full
        self.refresh()

    def refresh(self):
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        self.active_witnessess = self.blockchain.rpc.get_active_witnesses()["witnesses"]
        self.schedule = self.blockchain.rpc.get_witness_schedule()
        try:
            self.witness_count = self.blockchain.rpc.get_witness_count()
        except (RPCError, NoMethodWithName):
            try:
                # Some nodes expose the method only on condenser_api; fall back to that.
                self.witness_count = self.blockchain.rpc.get_witness_count(api="condenser_api")
            except Exception:
                # As a last resort, derive the count from the active witness set.
                self.witness_count = len(self.active_witnessess)
        self.current_witness = self.blockchain.get_dynamic_global_properties(use_stored_data=False)[
            "current_witness"
        ]
        self.identifier = ""
        super().__init__(
            [
                Witness(x, lazy=self.lazy, full=self.full, blockchain_instance=self.blockchain)
                for x in self.active_witnessess
            ]
        )


class WitnessesVotedByAccount(WitnessesObject):
    """Obtain a list of witnesses which have been voted by an account

    :param str account: Account name
    :param nectar.nectar.nectar blockchain_instance: nectar instance to use when
        accessing the RPC

    .. code-block:: python

       >>> from nectar.witness import WitnessesVotedByAccount
       >>> WitnessesVotedByAccount("gtg")
       <WitnessesVotedByAccount gtg>

    """

    def __init__(self, account, lazy=False, full=True, blockchain_instance=None):
        """
        Initialize a WitnessesVotedByAccount collection for the given account.

        Resolves the provided account to a full Account object, reads the list of witnesses that the account voted for, and populates the list with Witness objects created with the specified loading flags. If the account has no witness votes recorded the constructor returns early and the collection remains empty. The instance identifier is set to the account name.

        Parameters:
            account (str|Account): Account name or Account-like object to inspect for witness votes.
            lazy (bool): If True, create Witness objects in lazy-loading mode. Defaults to False.
            full (bool): If True, request full witness data when constructing Witness objects. Defaults to True.

        Note:
            The blockchain instance is taken from the optional `blockchain_instance` argument or the shared default; it is not documented here as a parameter.
        """
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.account = Account(account, full=True, blockchain_instance=self.blockchain)
        account_name = self.account["name"]
        self.identifier = account_name
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        if "witnesses_voted_for" not in self.account:
            return
        limit = self.account["witnesses_voted_for"]
        witnessess_dict = self.blockchain.rpc.list_witness_votes(
            {"start": [account_name], "limit": limit, "order": "by_account_witness"}
        )["votes"]
        witnessess = []
        for w in witnessess_dict:
            witnessess.append(w["witness"])

        super().__init__(
            [
                Witness(x, lazy=lazy, full=full, blockchain_instance=self.blockchain)
                for x in witnessess
            ]
        )


class WitnessesRankedByVote(WitnessesObject):
    """Obtain a list of witnesses ranked by Vote

    :param str from_account: Witness name from which the lists starts (default = "")
    :param int limit: Limits the number of shown witnesses (default = 100)
    :param nectar.nectar.nectar blockchain_instance: nectar instance to use when
        accessing the RPC

    .. code-block:: python

       >>> from nectar.witness import WitnessesRankedByVote
       >>> WitnessesRankedByVote(limit=100)
       <WitnessesRankedByVote >

    """

    def __init__(
        self,
        from_account="",
        limit=100,
        lazy=False,
        full=False,
        blockchain_instance=None,
    ):
        """
        Initialize a list of witnesses ranked by vote, with optional pagination.

        Builds a WitnessesRankedByVote list starting at `from_account` and returning up to `limit`
        entries. The constructor transparently pages RPC calls when the requested `limit`
        exceeds the per-call query limit, handles RPC calls and wraps returned witness entries as Witness objects.

        Parameters:
            from_account (str): Account name to start ranking from (inclusive). When empty, ranking starts from the top.
            limit (int): Maximum number of witnesses to return.
            lazy (bool): If True, create Witness objects in lazy-loading mode.
            full (bool): If True, fully load each Witness on creation.

        Notes:
            - `blockchain_instance` is taken from the shared instance when not provided.
            - The method uses different RPC endpoints depending on the node configuration
              (appbase vs non-appbase, and condenser mode) and automatically pages results
              to satisfy `limit`.
            - Returns early (no list created) if no witnesses are found.
        """
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        witnessList = []
        last_limit = limit
        self.identifier = ""
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        query_limit = 1000

        if from_account == "":
            last_account = None
        else:
            last_account = Witness(from_account, blockchain_instance=self.blockchain)["votes"]
        if limit > query_limit:
            while last_limit > query_limit:
                tmpList = WitnessesRankedByVote(last_account, query_limit)
                if last_limit < limit:
                    witnessList.extend(tmpList[1:])
                    last_limit -= query_limit - 1
                else:
                    witnessList.extend(tmpList)
                    last_limit -= query_limit

                last_account = witnessList[-1]["votes"]
        if last_limit < limit:
            last_limit += 1
        witnessess = self.blockchain.rpc.list_witnesses(
            {"start": [0, last_account], "limit": last_limit, "order": "by_vote_name"}
        )["witnesses"]

        # self.witness_count = len(self.voted_witnessess)
        if last_limit < limit:
            witnessess = witnessess[1:]
        if len(witnessess) > 0:
            for x in witnessess:
                witnessList.append(
                    Witness(x, lazy=lazy, full=full, blockchain_instance=self.blockchain)
                )
        if len(witnessList) == 0:
            return
        super().__init__(witnessList)


class ListWitnesses(WitnessesObject):
    """List witnesses ranked by name

    :param str from_account: Witness name from which the list starts (default = "")
    :param int limit: Limits the number of shown witnesses (default = 100)
    :param nectar.nectar.nectar blockchain_instance: nectar instance to use when
        accessing the RPC

    .. code-block:: python

       >>> from nectar.witness import ListWitnesses
       >>> ListWitnesses(from_account="gtg", limit=100)
       <ListWitnesses gtg>

    """

    def __init__(
        self,
        from_account="",
        limit=100,
        lazy=False,
        full=False,
        blockchain_instance=None,
    ):
        """
        Initialize a ListWitnesses collection starting from a given account name.

        Creates a list of Witness objects beginning at `from_account` (lexicographic start)
        up to `limit` entries. If no witnesses are found the constructor returns early
        leaving the instance empty. The object uses the provided blockchain instance
        (or the shared default) to query the node and sets `identifier` to `from_account`.

        Parameters:
            from_account (str): Account name to start listing witnesses from (inclusive).
            limit (int): Maximum number of witness entries to retrieve.
            lazy (bool): If True, construct Witness objects in lazy mode (defer full data load).
            full (bool): If True, request full witness data when constructing Witness objects.
        """
        self.blockchain = blockchain_instance or shared_blockchain_instance()
        self.identifier = from_account
        self.blockchain.rpc.set_next_node_on_empty_reply(False)
        witnessess = self.blockchain.rpc.list_witnesses(
            {"start": from_account, "limit": limit, "order": "by_name"}
        )["witnesses"]
        if len(witnessess) == 0:
            return
        super().__init__(
            [
                Witness(x, lazy=lazy, full=full, blockchain_instance=self.blockchain)
                for x in witnessess
            ]
        )



import hashlib
import logging
import sys
from binascii import hexlify, unhexlify
from typing import Any

import scrypt
from Cryptodome.Cipher import AES

from .account import PrivateKey
from .base58 import Base58, base58decode

log = logging.getLogger(__name__)


class SaltException(Exception):
    pass


def _encrypt_xor(a: Any, b: bytes, aes: Any) -> bytes:
    """Returns encrypt(a ^ b)."""
    a = unhexlify("%0.32x" % (int((a), 16) ^ int(hexlify(b), 16)))
    return aes.encrypt(a)


def encrypt(privkey: Any, passphrase: str) -> Base58:
    """BIP0038 non-ec-multiply encryption. Returns BIP0038 encrypted privkey.

    :param privkey: Private key
    :type privkey: Base58
    :param str passphrase: UTF-8 encoded passphrase for encryption
    :return: BIP0038 non-ec-multiply encrypted wif key
    :rtype: Base58

    """
    if isinstance(privkey, str):
        privkey = PrivateKey(privkey)
    else:
        privkey = PrivateKey(repr(privkey))

    privkeyhex = repr(privkey)  # hex
    addr = format(privkey.bitcoin.address, "BTC")
    a = bytes(addr, "ascii")
    salt = hashlib.sha256(hashlib.sha256(a).digest()).digest()[0:4]
    if sys.version < "3":
        if isinstance(passphrase, str):
            passphrase_bytes = passphrase.encode("utf-8")
        else:
            passphrase_bytes = passphrase
    else:
        passphrase_bytes = passphrase.encode("utf-8") if isinstance(passphrase, str) else passphrase

    key = scrypt.hash(passphrase_bytes, salt, 16384, 8, 8)
    (derived_half1, derived_half2) = (key[:32], key[32:])
    aes = AES.new(derived_half2, AES.MODE_ECB)
    encrypted_half1 = _encrypt_xor(privkeyhex[:32], derived_half1[:16], aes)
    encrypted_half2 = _encrypt_xor(privkeyhex[32:], derived_half1[16:], aes)
    " flag byte is forced 0xc0 because Graphene only uses compressed keys "
    payload = b"\x01" + b"\x42" + b"\xc0" + salt + encrypted_half1 + encrypted_half2
    " Checksum "
    checksum = hashlib.sha256(hashlib.sha256(payload).digest()).digest()[:4]
    privatkey = hexlify(payload + checksum).decode("ascii")
    return Base58(privatkey)


def decrypt(encrypted_privkey: Any, passphrase: str) -> Base58:
    """BIP0038 non-ec-multiply decryption. Returns WIF privkey.

    :param Base58 encrypted_privkey: Private key
    :param str passphrase: UTF-8 encoded passphrase for decryption
    :return: BIP0038 non-ec-multiply decrypted key
    :rtype: Base58
    :raises SaltException: if checksum verification failed (e.g. wrong password)

    """

    d = unhexlify(base58decode(encrypted_privkey))
    d = d[2:]  # remove trailing 0x01 and 0x42
    flagbyte = d[0:1]  # get flag byte
    d = d[1:]  # get payload
    if not flagbyte == b"\xc0":
        raise AssertionError("Flagbyte has to be 0xc0")
    salt = d[0:4]
    d = d[4:-4]
    if sys.version < "3":
        if isinstance(passphrase, str):
            passphrase_bytes = passphrase.encode("utf-8")
        else:
            passphrase_bytes = passphrase
    else:
        passphrase_bytes = passphrase.encode("utf-8") if isinstance(passphrase, str) else passphrase
    key = scrypt.hash(passphrase_bytes, salt, 16384, 8, 8)
    derivedhalf1 = key[0:32]
    derivedhalf2 = key[32:64]
    encryptedhalf1 = d[0:16]
    encryptedhalf2 = d[16:32]
    aes = AES.new(derivedhalf2, AES.MODE_ECB)
    decryptedhalf2 = aes.decrypt(encryptedhalf2)
    decryptedhalf1 = aes.decrypt(encryptedhalf1)
    privraw = decryptedhalf1 + decryptedhalf2
    privraw = "%064x" % (int(hexlify(privraw), 16) ^ int(hexlify(derivedhalf1), 16))
    wif = Base58(privraw)
    """ Verify Salt """
    privkey = PrivateKey(format(wif, "wif"))
    addr = format(privkey.bitcoin.address, "BTC")
    a = bytes(addr, "ascii")
    saltverify = hashlib.sha256(hashlib.sha256(a).digest()).digest()[0:4]
    if saltverify != salt:
        raise SaltException("checksum verification failed! Password may be incorrect.")
    return wif



import unittest

from nectargraphenebase.account import PrivateKey
from nectargraphenebase.bip38 import decrypt, encrypt


class Testcases(unittest.TestCase):
    def test_encrypt(self):
        self.assertEqual(
            [
                format(
                    encrypt(
                        PrivateKey("5HqUkGuo62BfcJU5vNhTXKJRXuUi9QSE6jp8C3uBJ2BVHtB8WSd"),
                        "TestingOneTwoThree",
                    ),
                    "encwif",
                ),
                format(
                    encrypt(
                        PrivateKey("5KN7MzqK5wt2TP1fQCYyHBtDrXdJuXbUzm4A9rKAteGu3Qi5CVR"),
                        "TestingOneTwoThree",
                    ),
                    "encwif",
                ),
                format(
                    encrypt(
                        PrivateKey("5HtasZ6ofTHP6HCwTqTkLDuLQisYPah7aUnSKfC7h4hMUVw2gi5"), "Satoshi"
                    ),
                    "encwif",
                ),
            ],
            [
                "6PRN5mjUTtud6fUXbJXezfn6oABoSr6GSLjMbrGXRZxSUcxThxsUW8epQi",
                "6PRVWUbkzzsbcVac2qwfssoUJAN1Xhrg6bNk8J7Nzm5H7kxEbn2Nh2ZoGg",
                "6PRNFFkZc2NZ6dJqFfhRoFNMR9Lnyj7dYGrzdgXXVMXcxoKTePPX1dWByq",
            ],
        )

    def test_deencrypt(self):
        self.assertEqual(
            [
                format(
                    decrypt(
                        "6PRN5mjUTtud6fUXbJXezfn6oABoSr6GSLjMbrGXRZxSUcxThxsUW8epQi",
                        "TestingOneTwoThree",
                    ),
                    "wif",
                ),
                format(
                    decrypt(
                        "6PRVWUbkzzsbcVac2qwfssoUJAN1Xhrg6bNk8J7Nzm5H7kxEbn2Nh2ZoGg",
                        "TestingOneTwoThree",
                    ),
                    "wif",
                ),
                format(
                    decrypt(
                        "6PRNFFkZc2NZ6dJqFfhRoFNMR9Lnyj7dYGrzdgXXVMXcxoKTePPX1dWByq", "Satoshi"
                    ),
                    "wif",
                ),
            ],
            [
                "5HqUkGuo62BfcJU5vNhTXKJRXuUi9QSE6jp8C3uBJ2BVHtB8WSd",
                "5KN7MzqK5wt2TP1fQCYyHBtDrXdJuXbUzm4A9rKAteGu3Qi5CVR",
                "5HtasZ6ofTHP6HCwTqTkLDuLQisYPah7aUnSKfC7h4hMUVw2gi5",
            ],
        )


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import datetime

from nectar import Hive, exceptions
from nectar.block import Block, BlockHeader
from nectar.instance import set_shared_blockchain_instance

from .nodes import get_hive_nodes

wif = "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"


class Testcases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        Set up class-level Hive blockchain test fixture.

        Creates a Hive instance (nobroadcast=True, provided active key, retries=10), stores it on the class as `bts`, sets `test_block_id` to 19273700, registers the instance as the shared blockchain instance, and sets the default account to "test".
        """
        cls.bts = Hive(
            node=get_hive_nodes(), nobroadcast=True, keys={"active": wif}, num_retries=10
        )
        cls.test_block_id = 19273700
        # from getpass import getpass
        # self.bts.wallet.unlock(getpass())
        set_shared_blockchain_instance(cls.bts)
        cls.bts.set_default_account("test")

    def test_block(self):
        bts = self.bts
        test_block_id = self.test_block_id
        block = Block(test_block_id, blockchain_instance=bts)
        self.assertEqual(block.identifier, test_block_id)
        self.assertTrue(isinstance(block.time(), datetime))
        self.assertTrue(isinstance(block, dict))

        self.assertTrue(len(block.operations))
        self.assertTrue(isinstance(block.ops_statistics(), dict))

        block2 = Block(test_block_id + 1, blockchain_instance=bts)
        self.assertTrue(block2.time() > block.time())
        with self.assertRaises(exceptions.BlockDoesNotExistsException):
            Block(0, blockchain_instance=bts)

    def test_block_only_ops(self):
        bts = self.bts
        test_block_id = self.test_block_id
        block = Block(test_block_id, only_ops=True, blockchain_instance=bts)
        self.assertEqual(block.identifier, test_block_id)
        self.assertTrue(isinstance(block.time(), datetime))
        self.assertTrue(isinstance(block, dict))

        self.assertTrue(len(block.operations))
        self.assertTrue(isinstance(block.ops_statistics(), dict))

        block2 = Block(test_block_id + 1, blockchain_instance=bts)
        self.assertTrue(block2.time() > block.time())
        with self.assertRaises(exceptions.BlockDoesNotExistsException):
            Block(0, blockchain_instance=bts)

    def test_block_header(self):
        bts = self.bts
        test_block_id = self.test_block_id
        block = BlockHeader(test_block_id, blockchain_instance=bts)
        self.assertEqual(block.identifier, test_block_id)
        self.assertTrue(isinstance(block.time(), datetime))
        self.assertTrue(isinstance(block, dict))

        block2 = BlockHeader(test_block_id + 1, blockchain_instance=bts)
        self.assertTrue(block2.time() > block.time())
        with self.assertRaises(exceptions.BlockDoesNotExistsException):
            BlockHeader(0, blockchain_instance=bts)

    def test_export(self):
        bts = self.bts
        block_num = 2000000

        block = bts.rpc.get_block({"block_num": block_num}, api="block_api")
        if block and "block" in block:
            block = block["block"]

        b = Block(block_num, blockchain_instance=bts)
        keys = list(block.keys())
        json_content = b.json()

        for k in keys:
            if k not in "json_metadata":
                if isinstance(block[k], dict) and isinstance(json_content[k], list):
                    self.assertEqual(list(block[k].values()), json_content[k])
                else:
                    self.assertEqual(block[k], json_content[k])

        block = bts.rpc.get_block_header({"block_num": block_num}, api="block_api")
        if "header" in block:
            block = block["header"]

        b = BlockHeader(block_num, blockchain_instance=bts)
        keys = list(block.keys())
        json_content = b.json()

        for k in keys:
            if k not in "json_metadata":
                if isinstance(block[k], dict) and isinstance(json_content[k], list):
                    self.assertEqual(list(block[k].values()), json_content[k])
                else:
                    self.assertEqual(block[k], json_content[k])



import logging

from nectar import Hive as Hive
from nectar.transactionbuilder import TransactionBuilder
from nectarbase import operations

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

# example wif
wif = "5KQwrPbwdL6PhXujxW37FSSQZ1JiwsST4cqQzDeyXtP79zkvFD3"


if __name__ == "__main__":
    hv_online = Hive()
    trx_builder = TransactionBuilder(blockchain_instance=hv_online)
    ref_block_num, ref_block_prefix = trx_builder.get_block_params()
    print("ref_block_num %d - ref_block_prefix %d" % (ref_block_num, ref_block_prefix))

    hv = Hive(offline=True)

    op = operations.Transfer(
        {"from": "thecrazygm", "to": "thecrazygm", "amount": "0.001 HBD", "memo": ""}
    )
    tb = TransactionBuilder(blockchain_instance=hv)

    tb.appendOps([op])
    tb.appendWif(wif)
    tb.constructTx(ref_block_num=ref_block_num, ref_block_prefix=ref_block_prefix)
    tx = tb.sign(reconstruct_tx=False)
    print(tx.json())



import hashlib
import logging
import string
from binascii import hexlify, unhexlify
from typing import Any, Optional, Union

from .prefix import Prefix

log = logging.getLogger(__name__)


class Base58(Prefix):
    """Base58 base class

    This class serves as an abstraction layer to deal with base58 encoded
    strings and their corresponding hex and binary representation throughout the
    library.

    :param data: Data to initialize object, e.g. pubkey data, address data, ...
    :type data: hex, wif, bip38 encrypted wif, base58 string
    :param str prefix: Prefix to use for Address/PubKey strings (defaults to ``GPH``)
    :return: Base58 object initialized with ``data``
    :rtype: Base58
    :raises ValueError: if data cannot be decoded

    * ``bytes(Base58)``: Returns the raw data
    * ``str(Base58)``:   Returns the readable ``Base58CheckEncoded`` data.
    * ``repr(Base58)``:  Gives the hex representation of the data.
    * ``format(Base58,_format)``: Formats the instance according to ``_format``

        * ``"btc"``: prefixed with ``0x80``. Yields a valid btc address
        * ``"wif"``: prefixed with ``0x00``. Yields a valid wif key
        * ``"bts"``: prefixed with ``BTS``
        * etc.

    """

    def __init__(self, data: Any, prefix: Optional[str] = None) -> None:
        self.set_prefix(prefix)
        if isinstance(data, Base58):
            data = repr(data)
        if all(c in string.hexdigits for c in data):
            self._hex = data
        elif data[0] == "5" or data[0] == "6":
            self._hex = base58CheckDecode(data)
        elif data[0] == "K" or data[0] == "L":
            self._hex = base58CheckDecode(data)[:-2]
        elif data[: len(self.prefix)] == self.prefix:
            self._hex = gphBase58CheckDecode(data[len(self.prefix) :])
        else:
            raise ValueError("Error loading Base58 object")

    def __format__(self, _format: str) -> str:
        """Format output according to argument _format (wif,btc,...)

        :param str _format: Format to use
        :return: formatted data according to _format
        :rtype: str

        """
        if _format.upper() == "WIF":
            return base58CheckEncode(0x80, self._hex)
        elif _format.upper() == "ENCWIF":
            return base58encode(self._hex)
        elif _format.upper() == "BTC":
            return base58CheckEncode(0x00, self._hex)
        else:
            return _format.upper() + str(self)

    def __repr__(self) -> str:
        """Returns hex value of object

        :return: Hex string of instance's data
        :rtype: hex string
        """
        return self._hex

    def __str__(self) -> str:
        """Return graphene-base58CheckEncoded string of data

        :return: Base58 encoded data
        :rtype: str
        """
        return gphBase58CheckEncode(self._hex)

    def __bytes__(self) -> bytes:
        """Return raw bytes

        :return: Raw bytes of instance
        :rtype: bytes

        """
        return unhexlify(self._hex)


# https://github.com/tochev/python3-cryptocoins/raw/master/cryptocoins/base58.py
BASE58_ALPHABET = b"123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"


def base58decode(base58_str: str) -> str:
    base58_text = bytes(base58_str, "ascii")
    n = 0
    leading_zeroes_count = 0
    for b in base58_text:
        # b is always int in Python 3 when iterating bytes
        n = n * 58 + BASE58_ALPHABET.find(bytes([b]))
        if n == 0:
            leading_zeroes_count += 1
    res = bytearray()
    while n >= 256:
        div, mod = divmod(n, 256)
        res.insert(0, mod)
        n = div
    else:
        res.insert(0, n)
    return hexlify(bytearray(1) * leading_zeroes_count + res).decode("ascii")


def base58encode(hexstring: str) -> str:
    # Handle odd-length hex strings by padding with leading zero
    if len(hexstring) % 2 == 1:
        hexstring = "0" + hexstring
    byteseq = bytes(unhexlify(bytes(hexstring, "ascii")))
    n = 0
    leading_zeroes_count = 0
    for c in byteseq:
        n = n * 256 + c
        if n == 0:
            leading_zeroes_count += 1
    res = bytearray()
    while n >= 58:
        div, mod = divmod(n, 58)
        res.insert(0, BASE58_ALPHABET[mod])
        n = div
    else:
        res.insert(0, BASE58_ALPHABET[n])
    return (BASE58_ALPHABET[0:1] * leading_zeroes_count + res).decode("ascii")


def ripemd160(s: Union[str, bytes]) -> bytes:
    ripemd160 = hashlib.new("ripemd160")
    ripemd160.update(unhexlify(s))
    return ripemd160.digest()


def doublesha256(s: Union[str, bytes]) -> bytes:
    if isinstance(s, str):
        # Handle odd-length hex strings by padding with leading zero
        if len(s) % 2 == 1:
            s = "0" + s
        s = unhexlify(s)
    return hashlib.sha256(hashlib.sha256(s).digest()).digest()


def b58encode(v: str) -> str:
    return base58encode(v)


def b58decode(v: str) -> str:
    return base58decode(v)


def base58CheckEncode(version: int, payload: str) -> str:
    if isinstance(version, str):
        s = version + payload
    else:
        s = ("%.2x" % version) + payload
    checksum = doublesha256(s)[:4]
    result = s + hexlify(checksum).decode("ascii")
    return base58encode(result)


def base58CheckDecode(s: str, skip_first_bytes: bool = True) -> str:
    s_bytes = unhexlify(base58decode(s))
    dec = hexlify(s_bytes[:-4]).decode("ascii")
    checksum = doublesha256(dec)[:4]
    if not (s_bytes[-4:] == checksum):
        raise AssertionError()
    if skip_first_bytes:
        return dec[2:]
    else:
        return dec


def gphBase58CheckEncode(s: str) -> str:
    checksum = ripemd160(s)[:4]
    result = s + hexlify(checksum).decode("ascii")
    return base58encode(result)


def gphBase58CheckDecode(s: str) -> str:
    s_bytes = unhexlify(base58decode(s))
    dec = hexlify(s_bytes[:-4]).decode("ascii")
    checksum = ripemd160(dec)[:4]
    if not (s_bytes[-4:] == checksum):
        raise AssertionError()
    return dec



import logging
import time
from datetime import datetime, timedelta

from nectar.block import Block
from nectar.blockchain import Blockchain
from nectar.hive import Hive
from nectar.nodelist import NodeList
from nectar.utils import formatTimedelta

log = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


if __name__ == "__main__":
    node_setup = 1
    how_many_hours = 1
    nodes = NodeList()
    if node_setup == 0:
        hv = Hive(node=nodes.get_nodes(normal=True, https=True), num_retries=10)
        max_batch_size = None
        threading = False
        thread_num = 8
    elif node_setup == 1:
        hv = Hive(node=nodes.get_nodes(normal=True, https=True), num_retries=10)
        max_batch_size = None
        threading = True
        thread_num = 16
    elif node_setup == 2:
        hv = Hive(node=nodes.get_nodes(appbase=False, https=True), num_retries=10)
        max_batch_size = None
        threading = True
        thread_num = 16
    blockchain = Blockchain(blockchain_instance=hv)
    last_block_id = 19273700
    last_block = Block(last_block_id, blockchain_instance=hv)
    startTime = datetime.now()

    stopTime = last_block.time() + timedelta(seconds=how_many_hours * 60 * 60)
    ltime = time.time()
    cnt = 0
    total_transaction = 0

    start_time = time.time()
    last_node = hv.rpc.url
    print("Current node:", last_node)
    for entry in blockchain.blocks(
        start=last_block_id,
        max_batch_size=max_batch_size,
        threading=threading,
        thread_num=thread_num,
        thread_limit=1200,
    ):
        block_no = entry.identifier
        if "block" in entry:
            trxs = entry["block"]["transactions"]
        else:
            trxs = entry["transactions"]

        for tx in trxs:
            for op in tx["operations"]:
                total_transaction += 1
        if "block" in entry:
            block_time = entry["block"]["timestamp"]
        else:
            block_time = entry["timestamp"]

        if block_time > stopTime:
            total_duration = formatTimedelta(datetime.now() - startTime)
            last_block_id = block_no
            avtran = total_transaction / (last_block_id - 19273700)
            print(
                "* HOUR mark: Processed %d blockchain hours in %s"
                % (how_many_hours, total_duration)
            )
            print(
                "* Blocks %d, Transactions %d (Avg. per Block %f)"
                % ((last_block_id - 19273700), total_transaction, avtran)
            )
            break

        if block_no != last_block_id:
            cnt += 1
            last_block_id = block_no
            if last_block_id % 100 == 0:
                now = time.time()
                duration = now - ltime
                total_duration = now - start_time
                speed = int(100000.0 / duration) * 1.0 / 1000
                avspeed = int((last_block_id - 19273700) * 1000 / total_duration) * 1.0 / 1000
                avtran = total_transaction / (last_block_id - 19273700)
                ltime = now
                if last_node != hv.rpc.url:
                    last_node = hv.rpc.url
                    print("Current node:", last_node)
                print(
                    "* 100 blocks processed in %.2f seconds. Speed %.2f. Avg: %.2f. Avg.Trans:"
                    "%.2f Count: %d Block minutes: %d"
                    % (duration, speed, avspeed, avtran, cnt, cnt * 3 / 60)
                )
